from .et import ETOut
from .mulch import MulchOut
from .plantgro import PlantGroOut
from .soilni import SoilNiOut
from .soiltemp import SoilTempOut
from .soilwat import SoilWat
from .summary import SummaryOut



from tradssat.tmpl import InpFile
from .wth_vars import vars_


class WTHFile(InpFile):
    """
    Reader for DSSAT weather (.WTH and .WTG) files.
    """

    ext = ['.WTH', '.WTG']

    def _get_var_info(self):
        return vars_



import json
import os
from copy import deepcopy
from json import JSONDecodeError
from tempfile import NamedTemporaryFile
from warnings import warn

import numpy as np
import numpy.testing as npt


def find_files(inp_class, folder):
    l_files = []
    for root, dirs, files in os.walk(folder):
        for f in files:
            if inp_class.matches_file(f):
                l_files.append(os.path.join(root, f))

    return l_files


def get_ref(file, default):
    path, file = os.path.split(file)
    file_name = os.path.splitext(file)[0]
    ref_file = os.path.join(path, '_ref_{}.json'.format(file))
    try:
        return read_json(ref_file)
    except (FileNotFoundError, JSONDecodeError):
        write_json(default, ref_file)
        warn('Reference generated for tmpl "{}". Check that its contents are correct.'.format(file_name))
        return default


def test_read(inp_class, folder, testcase):
    files = find_files(inp_class, folder)

    for f in files:
        with testcase.subTest(os.path.split(f)[1]):
            dict_vars = inp_class(f).to_dict()

            ref = get_ref(f, dict_vars)
            _test_dicts_equal(tc=testcase, act=dict_vars, ref=ref, f=f)


def test_write(inp_class, folder, testcase):
    ext = inp_class.ext
    files = find_files(inp_class, folder)

    for f in files:
        with testcase.subTest(os.path.split(f)[1]):
            if isinstance(ext, list):
                ext = ext[0]

            name = os.path.split(f)[1]

            temp_file = NamedTemporaryFile(prefix=name, suffix=ext).name
            inp_file_obj = inp_class(f)
            inp_file_obj.write(temp_file)

            new_file_obj = inp_class(temp_file)

            _test_dicts_equal(testcase, act=new_file_obj.to_dict(), ref=inp_file_obj.to_dict(), f=f)


def _test_dicts_equal(tc, act, ref, f, keys=None):
    if keys is None:
        keys = []
    if isinstance(act, dict):
        tc.assertListEqual(list(act), list(ref), msg='Dictionary keys are not equal in {}'.format(_f_loc(f, keys)))
        for k, v in act.items():
            keys.append(k)
            _test_dicts_equal(tc=tc, act=v, ref=ref[k], f=f, keys=keys)
            keys[:] = keys[:-1]
    elif isinstance(act, list):
        tc.assertEqual(len(act), len(ref), msg='List lengths not equal in {}'.format(_f_loc(f, keys)))
        for i, (a, r) in enumerate(zip(act, ref)):
            keys.append(i)
            _test_dicts_equal(tc=tc, act=a, ref=r, f=f, keys=keys)
            keys[:] = keys[:-1]
    elif isinstance(act, np.ndarray):
        if not len(act) == len(ref) == 0:  # skip empty arrays (which may have inferred different types)
            npt.assert_equal(actual=act, desired=ref, err_msg=_f_loc(f, keys))


def _f_loc(f, keys):
    return '{}\n\t{}'.format(f, ': '.join(str(k) for k in keys))


def write_json(obj, file):
    obj = deepcopy(obj)
    jsonify(obj)
    with open(file, 'w', encoding='utf-8') as f:
        json.dump(obj, f, indent=2, ensure_ascii=False)


def read_json(file):
    with open(file, 'r', encoding='utf-8') as f:
        return json.load(f)


def jsonify(d):
    for k, v in d.items():
        if isinstance(v, np.ndarray):
            d[k] = v.tolist()
        elif isinstance(v, dict):
            jsonify(v)
        elif isinstance(v, list):
            for i in v:
                jsonify(i)



import os
import unittest

from tests.utils import test_read, test_write
from tradssat.out import SoilTempOut, SoilNiOut, SummaryOut, PlantGroOut, ETOut, SoilWat, MulchOut
from tradssat import SoilFile, WTHFile, MTHFile, ExpFile, CULFile, ECOFile

rsrcs = os.path.join(os.path.split(__file__)[0], 'rsrc/mock_DSSAT')
input_classes = [SoilFile, WTHFile, MTHFile, ExpFile, CULFile, ECOFile]

rsrcs_out = os.path.join(os.path.split(__file__)[0], 'rsrc/mock_DSSAT/Out')
output_classes = [PlantGroOut, SoilNiOut, SoilTempOut, SoilWat, MulchOut, ETOut]
final_out_classes = [SummaryOut]


# Inputs must be read and written
class TestInputs(unittest.TestCase):
    def test_read(self):
        for inp_class in input_classes:
            with self.subTest(inp_class.__name__):
                test_read(inp_class, folder=rsrcs, testcase=self)

    def test_write(self):
        for inp_class in input_classes:
            with self.subTest(inp_class.__name__):
                test_write(inp_class, rsrcs, testcase=self)


# Outputs are only read, not written
class TestOutputs(unittest.TestCase):
    def test_read(self):
        for out_class in output_classes:
            with self.subTest(out_class.__name__):
                test_read(out_class, folder=rsrcs_out, testcase=self)


class TestFinalOutputs(unittest.TestCase):
    def test_read(self):
        for final_out_class in final_out_classes:
            with self.subTest(final_out_class.__name__):
                test_read(final_out_class, folder=rsrcs_out, testcase=self)



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_CSYCA = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXP#', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar('PPS1', 5, 0, info='Photoperiod sensitivity for phase 1. (% drop for 10h pp.change)'),
    FloatVar('B01ND', 5, 0, info='Duration from branch 0 to branch 1 (ie.tier x,node number)'),
    FloatVar('B12ND', 5, 0, info='Duration from branch 1 to branch 2 (ie.tier x,node number)'),
    FloatVar(
        'HMPC', 5, 0,
        info='Harvest product moisture content (%) (currently this variable doesn\'t affect calibration)'
    ),
    FloatVar('LA1S', 5, 0, info='Area/leaf (cm2) of the first leaves when growing without stress.'),
    FloatVar('LAXS', 5, 0, info='Area/leaf at maximum area/leaf (cm2)'),
    FloatVar('SLAS', 5, 0, info='Specific leaf lamina area when crop growing without stress (cm2/g)'),
    FloatVar('LLIFA', 5, 0, info='Leaf life,from full expansion to start senescence (Thermal units)'),
    FloatVar('LPEFR', 5, 2, info='Leaf petiole fraction (fr of lamina+petiole)'),
    FloatVar('LNSLP', 5, 1, info='Slope for leaf production (0.8 low rate, 1.0 medium rate, 1.2 high rate)'),
    FloatVar('NODWT', 5, 1, info='Node weight for the first stem of the shoot before branching at 3400 ˚Cd'),
    FloatVar(
        'NODLT', 5, 0,
        info='Mean internode length (cm) for for the first stem of the shoot before branching when is lignified'
    )
}

eco_vars_CSYCA = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('PARUE', 5, 2, info='PAR conversion factor,standard (g dry matter/MJ)'),
    FloatVar('TBLSZ', 5, 1, info='Base temperature for leaf development (˚C)'),
    FloatVar('BR1FX', 5, 1, info='Branch number per fork at fork 1, maximum (#)'),
    FloatVar('BR2FX', 5, 1, info='Branch number per fork at fork 2, maximum (#)'),
    FloatVar('BR3FX', 5, 1, info='Branch number per fork at fork 3, maximum (#)'),
    FloatVar('BR4FX', 5, 1, info='Branch number per fork at fork 4, maximum (#)'),
    FloatVar('DUSRI', 5, 1, info='Degree days at which storage root number determined (oC.d)'),
    FloatVar('SRN%S', 5, 2, info='Storage root standard N concentration (% dm)'),
    FloatVar('KCAN', 5, 2, info='PAR extinction coefficient (#)')
}



import os

import numpy as np

from .file import File
from .input import InpFile
from .vals import ValueSubSection


class OutFile(File):
    """
    Parent class for (nearly all) DSSAT output files.
    """

    filename = None  # type: str

    def __init__(self, folder):
        if os.path.isdir(folder):
            file = os.path.join(folder, self.filename)
        else:
            file = folder

        super().__init__(file)

    def _process_section_header(self, lines):

        section_name = lines[0][1:]
        self._values.add_section(section_name)
        i = 0
        for l in lines:
            if 'TREATMENT' in l:
                trt_no = l.split(' : ')[0].split()[1].strip()
                subsect = ValueSubSection(['TREATMENT'], l_vals=[np.array([trt_no], dtype=int)])

                self._values[section_name].set_header_vars(subsect)

            elif '@' in l:
                break
            i += 1

        return section_name, lines[i:]

    def _get_var_info(self):
        raise NotImplementedError

    @classmethod
    def matches_file(cls, file):
        """
        Checks whether a given file can be read by this class.
        
        Parameters
        ----------
        file: str
           The filename or full path to be read.
        
        Returns
        -------
        bool
           ``True`` if the file matches; ``False`` otherwise.
        """

        fname = os.path.split(file)[1]
        return fname.lower() == cls.filename.lower()


class FinalOutFile(InpFile):
    filename = None  # type: str

    def __init__(self, folder):
        if os.path.isdir(folder):
            file = os.path.join(folder, self.filename)
        else:
            file = folder

        super().__init__(file)

    @classmethod
    def matches_file(cls, file):
        fname = os.path.split(file)[1]
        return fname.lower() == cls.filename.lower()

    def _get_var_info(self):
        raise NotImplementedError



import numpy as np

from .utils import _name_matches

CODE_MISS = -99


class Variable(object):
    """
    Parent class for all variable types.
    """
    type_ = None

    def __init__(self, name, size, spc, header_fill, float_r, miss, sect, info):
        self.name = name
        self.size = size
        self.spc = spc

        self.fill = header_fill
        self.float_r = float_r
        self.miss = miss

        self.info = info
        self.sect = sect

    def write(self, val=None):
        fill = self.fill if val is None else ' '
        if val is None:
            txt = str(self)
        else:
            if val == CODE_MISS:
                val = self.miss
            txt = self._write(val)

        if self.float_r:
            return ' ' * self.spc + txt.ljust(self.size, fill)
        return ' ' * self.spc + txt.rjust(self.size, fill)

    def check_val(self, val):
        raise NotImplementedError

    def _write(self, val):
        raise NotImplementedError

    def __str__(self):
        return str(self.name)


class CharacterVar(Variable):
    """
    Character (string) variable.
    """
    type_ = str

    def __init__(self, name, size, spc=1, sect=None, header_fill=' ', miss='-99', info=''):
        super().__init__(name, size, spc, header_fill, float_r=False, miss=miss, sect=sect, info=info)

    def check_val(self, val):
        if isinstance(val, str):
            if len(val) > self.size:
                raise ValueError(
                    'Value "{val}" is longer than {sz} characters for variable {name}.'.format(
                        val=val, sz=self.size, name=self
                    )
                )

    def _write(self, val):
        return val


class NumericVar(Variable):
    """
    Parent class for numeric variable types.
    """

    def __init__(self, name, size, lims, spc, header_fill, miss, sect, info):

        super().__init__(name, size, spc, header_fill, sect=sect, float_r=True, miss=miss, info=info)

        if lims is None:
            lims = (-np.inf, np.inf)
        lims = (-np.inf if lims[0] is None else lims[0], np.inf if lims[1] is None else lims[1])

        self.lims = lims

    def check_val(self, val):
        val = np.array(val)
        out = np.where(np.logical_or(np.less(val, self.lims[0]), np.greater(val, self.lims[1])))
        if out[0].size > 0:
            vals_out = val[out]
            raise ValueError(
                'Value {val} is not in range {rng} for variable {name}.'.format(val=vals_out, name=self, rng=self.lims)
            )

    def _write(self, val):
        raise NotImplementedError


class FloatVar(NumericVar):
    """
    Floating point (decimal) variable.
    """
    type_ = float

    def __init__(self, name, size, dec, lims=None, spc=1, sect=None, header_fill=' ', miss='-99', info=''):
        super().__init__(name, size, lims, spc, header_fill, miss=miss, sect=sect, info=info)
        self.dec = dec

    def _write(self, val):
        if val == self.miss:
            return '-99'  # to avoid size overflow on small-sized variables with decimals
        # todo: clean
        txt_0 = str(val)
        space_req = len(txt_0.split('.')[0]) + 1
        if txt_0.startswith('0') or txt_0.startswith('-0'):
            space_req -= 1

        dec = min(self.dec, max(0, self.size - space_req))

        txt = '{:{sz}.{dec}f}'.format(val, sz=self.size, dec=dec)
        if txt[0] == '0':
            txt = txt[1:]
        elif txt[:2] == '-0':
            txt = '-{}'.format(txt[2:])
        return txt


class IntegerVar(NumericVar):
    """
    Integer (whole number) variable.
    """
    type_ = int

    def __init__(self, name, size, lims=None, spc=1, sect=None, header_fill=' ', miss='-99', info=''):
        super().__init__(name, size, lims, spc, header_fill, miss=miss, sect=sect, info=info)

    def _write(self, val):
        if val == self.miss:
            return '{:{sz}}'.format(val, sz=self.size)  # to avoid problems with non-numeric missing value codes
        return '{:{sz}d}'.format(val, sz=self.size)


class VariableSet(object):
    """
    Organiser for the allowed variables of a DSSAT file type.
    """

    def __init__(self, l_vars):
        self._vars = l_vars

    def get_var(self, var, sect=None):

        try:
            return next(
                vr for vr in self._vars if str(vr) == var and _name_matches(vr.sect, sect, full=True)
            )
        except StopIteration:
            if sect is None:
                msg = 'Variable {var} does not exist.'.format(var=var)
            else:
                msg = 'Variable {var} does not exist in section {sect}.'.format(var=var, sect=sect)
            raise ValueError(msg)

    def variables(self):
        return self._vars

    def __contains__(self, item):
        return any(str(vr) == str(item) for vr in self._vars)

    def __iter__(self):
        for vr in self._vars:
            yield vr


class HeaderVariableSet(object):
    """
    Organiser for the allowed header variables of a DSSAT file type.
    """

    def __init__(self, d_vars):
        self._vars = d_vars

    def matches(self, header):
        for hd in self._vars:
            match = _name_matches(hd, header)
            if match is not None:
                return match

    def get_vars(self, header):
        return next(self._vars[hd] for hd in self._vars if _name_matches(hd, header))

    def get_var(self, var, sect=None):
        try:
            return next(
                vr for hd, vrs in self._vars.items() for vr in vrs if _name_matches(hd, sect) and var == str(vr)
            )
        except StopIteration:
            raise ValueError('Variable {var} does not exist.'.format(var=var))



import os

from tradssat.tmpl import InpFile
from .vars_.alfrm import cul_vars_ALFRM, eco_vars_ALFRM
from .vars_.bacer import cul_vars_BACER, eco_vars_BACER
from .vars_.bacrp import cul_vars_BACRP, eco_vars_BACRP
from .vars_.bhgro import cul_vars_BHGRO, eco_vars_BHGRO
from .vars_.bmfrm import cul_vars_BMFRM, eco_vars_BMFRM
from .vars_.bmgro import cul_vars_BMGRO, eco_vars_BMGRO
from .vars_.bngro import cul_vars_BNGRO, eco_vars_BNGRO
from .vars_.brfrm import cul_vars_BRFRM, eco_vars_BRFRM
from .vars_.brgro import cul_vars_BRGRO, eco_vars_BRGRO
from .vars_.bscer import cul_vars_BSCER, eco_vars_BSCER
from .vars_.cbgro import cul_vars_CBGRO, eco_vars_CBGRO
from .vars_.chgro import cul_vars_CHGRO, eco_vars_CHGRO
from .vars_.cngro import cul_vars_CNGRO, eco_vars_CNGRO
from .vars_.cogro import cul_vars_COGRO, eco_vars_COGRO
from .vars_.cpgro import cul_vars_CPGRO, eco_vars_CPGRO
from .vars_.cscas import cul_vars_CSCAS, eco_vars_CSCAS
from .vars_.csyca import cul_vars_CSYCA, eco_vars_CSYCA
from .vars_.fbgro import cul_vars_FBGRO, eco_vars_FBGRO
from .vars_.g0gro import cul_vars_G0GRO, eco_vars_G0GRO
from .vars_.gbgro import cul_vars_GBGRO, eco_vars_GBGRO
from .vars_.mlcer import cul_vars_MLCER, eco_vars_MLCER
from .vars_.mzcer import cul_vars_MZCER, eco_vars_MZCER
from .vars_.mzixm import cul_vars_MZIXM, eco_vars_MZIXM
from .vars_.pialo import cul_vars_PIALO
from .vars_.pngro import cul_vars_PNGRO, eco_vars_PNGRO
from .vars_.ppgro import cul_vars_PPGRO, eco_vars_PPGRO
from .vars_.prgro import cul_vars_PRGRO, eco_vars_PRGRO
from .vars_.ptsub import cul_vars_PTSUB, eco_vars_PTSUB
from .vars_.ricer import cul_vars_RICER
from .vars_.riorz import cul_vars_RIORZ
from .vars_.sbgro import cul_vars_SBGRO, eco_vars_SBGRO
from .vars_.sccan import cul_vars_SCCAN, eco_vars_SCCAN
from .vars_.sccsp import cul_vars_SCCSP
from .vars_.sfgro import cul_vars_SFGRO, eco_vars_SFGRO
from .vars_.sgcer import cul_vars_SGCER, eco_vars_SGCER
from .vars_.sugro import cul_vars_SUGRO, eco_vars_SUGRO
from .vars_.swcer import cul_vars_SWCER, eco_vars_SWCER
from .vars_.tmgro import cul_vars_TMGRO, eco_vars_TMGRO
from .vars_.tnaro import cul_vars_TNARO
from .vars_.traro import cul_vars_TRARO
from .vars_.vbgro import cul_vars_VBGRO, eco_vars_VBGRO
from .vars_.whaps import cul_vars_WHAPS, eco_vars_WHAPS
from .vars_.whcer import cul_vars_WHCER, eco_vars_WHCER
from .vars_.whcrp import cul_vars_WHCRP, eco_vars_WHCRP

vars_gen = {
    'ALFRM': {'.cul': cul_vars_ALFRM, '.eco': eco_vars_ALFRM},
    'BACER': {'.cul': cul_vars_BACER, '.eco': eco_vars_BACER},
    'BACRP': {'.cul': cul_vars_BACRP, '.eco': eco_vars_BACRP},
    'BHGRO': {'.cul': cul_vars_BHGRO, '.eco': eco_vars_BHGRO},
    'BMFRM': {'.cul': cul_vars_BMFRM, '.eco': eco_vars_BMFRM},
    'BMGRO': {'.cul': cul_vars_BMGRO, '.eco': eco_vars_BMGRO},
    'BNGRO': {'.cul': cul_vars_BNGRO, '.eco': eco_vars_BNGRO},
    'BRFRM': {'.cul': cul_vars_BRFRM, '.eco': eco_vars_BRFRM},
    'BRGRO': {'.cul': cul_vars_BRGRO, '.eco': eco_vars_BRGRO},
    'BSCER': {'.cul': cul_vars_BSCER, '.eco': eco_vars_BSCER},
    'CBGRO': {'.cul': cul_vars_CBGRO, '.eco': eco_vars_CBGRO},
    'CHGRO': {'.cul': cul_vars_CHGRO, '.eco': eco_vars_CHGRO},
    'CNGRO': {'.cul': cul_vars_CNGRO, '.eco': eco_vars_CNGRO},
    'COGRO': {'.cul': cul_vars_COGRO, '.eco': eco_vars_COGRO},
    'CPGRO': {'.cul': cul_vars_CPGRO, '.eco': eco_vars_CPGRO},
    'CSCAS': {'.cul': cul_vars_CSCAS, '.eco': eco_vars_CSCAS},
    'CSYCA': {'.cul': cul_vars_CSYCA, '.eco': eco_vars_CSYCA},
    'FBGRO': {'.cul': cul_vars_FBGRO, '.eco': eco_vars_FBGRO},
    'G0GRO': {'.cul': cul_vars_G0GRO, '.eco': eco_vars_G0GRO},
    'GBGRO': {'.cul': cul_vars_GBGRO, '.eco': eco_vars_GBGRO},
    'MLCER': {'.cul': cul_vars_MLCER, '.eco': eco_vars_MLCER},
    'MZCER': {'.cul': cul_vars_MZCER, '.eco': eco_vars_MZCER},
    'MZIXM': {'.cul': cul_vars_MZIXM, '.eco': eco_vars_MZIXM},
    'PIALO': {'.cul': cul_vars_PIALO},
    'PNGRO': {'.cul': cul_vars_PNGRO, '.eco': eco_vars_PNGRO},
    'PPGRO': {'.cul': cul_vars_PPGRO, '.eco': eco_vars_PPGRO},
    'PRGRO': {'.cul': cul_vars_PRGRO, '.eco': eco_vars_PRGRO},
    'PTSUB': {'.cul': cul_vars_PTSUB, '.eco': eco_vars_PTSUB},
    'RICER': {'.cul': cul_vars_RICER},
    'RIORZ': {'.cul': cul_vars_RIORZ},
    'SBGRO': {'.cul': cul_vars_SBGRO, '.eco': eco_vars_SBGRO},
    'SCCAN': {'.cul': cul_vars_SCCAN, '.eco': eco_vars_SCCAN},
    'SCCSP': {'.cul': cul_vars_SCCSP},
    'SFGRO': {'.cul': cul_vars_SFGRO, '.eco': eco_vars_SFGRO},
    'SGCER': {'.cul': cul_vars_SGCER, '.eco': eco_vars_SGCER},
    'SUGRO': {'.cul': cul_vars_SUGRO, '.eco': eco_vars_SUGRO},
    'SWCER': {'.cul': cul_vars_SWCER, '.eco': eco_vars_SWCER},
    'TNARO': {'.cul': cul_vars_TNARO},
    'TMGRO': {'.cul': cul_vars_TMGRO, '.eco': eco_vars_TMGRO},
    'TRARO': {'.cul': cul_vars_TRARO},
    'VBGRO': {'.cul': cul_vars_VBGRO, '.eco': eco_vars_VBGRO},
    'WHAPS': {'.cul': cul_vars_WHAPS, '.eco': eco_vars_WHAPS},
    'WHCER': {'.cul': cul_vars_WHCER, '.eco': eco_vars_WHCER},
    'WHCRP': {'.cul': cul_vars_WHCRP, '.eco': eco_vars_WHCRP},
}


class GenFile(InpFile):
    """
    Template for genotype input file readers.
    """
    ext = None  # type: str

    def _get_var_info(self):
        file = os.path.split(self.file)[1]

        for gen, d_gen in vars_gen.items():
            if file.startswith(gen):
                try:
                    return d_gen[self.ext.lower()]
                except KeyError:
                    break

        raise ValueError(
            'No {ext} variables defined for {f} cropping model.'.format(ext=self.ext, f=os.path.splitext(file)[0])
        )


class CULFile(GenFile):
    """
    Cultivar (.CUL) input file reader.
    """
    ext = '.CUL'


class ECOFile(GenFile):
    """
    Ecotype (.ECO) input file reader.
    """
    ext = '.ECO'



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class ETOut(OutFile):
    filename = 'ET.OUT'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 5, info='Day after start'),
    FloatVar('SRAA', 6, 2, info='Average solar radiation (MJ/(m2.day'),
    FloatVar('TMAXA', 6, 2, info='Average maximum air temperature (°C)'),
    FloatVar('TMINA', 6, 2, info='Average minimum air temperature (°C)'),
    FloatVar('EOAA', 6, 3, info='Average potential evapotranspiration (mm/d)'),
    FloatVar('EOPA', 6, 3, info=' Average potential transpiration (mm/d)'),
    FloatVar('EOSA', 6, 3, info='Average potential soil evaporation per day (mm/day)'),
    FloatVar('ETAA', 6, 3, info='Average evapotranspiration (mm/d)'),
    FloatVar('EPAA', 6, 0, info=' Average plant transpiration (mm/d)'),
    FloatVar('ESAA', 6, 3, info='Average soil evaporation (mm/d)'),
    FloatVar('EFAA', 6, 0, info='Average flood evaporation (mm/d)'),
    FloatVar('EMAA', 6, 3, info='Average mulch evaporation (mm/d)'),
    FloatVar('EOAC', 7, 2, info='Cumulative potential evapotranspiration (mm)'),
    FloatVar('ETAC', 7, 2, info='Cumulative evapotranspiration (mm)'),
    FloatVar('EPAC', 7, 0, info=' Cumulative transpiration (mm)'),
    FloatVar('ESAC', 7, 2, info='Cumulative soil evaporation (mm)'),
    FloatVar('EFAC', 7, 0, info='Cumulative floodwater evaporation (mm)'),
    FloatVar('EMAC', 7, 2, info='Cumulative mulch evaporation (mm)'),
}



import os
import re

import numpy as np

from tradssat.utils import detect_encod
from .vals import FileValueSet, ValueSubSection
from .var import VariableSet, CODE_MISS


class File(object):
    """
    Parent class for all file objects.
    """

    def __init__(self, file):
        """

        Parameters
        ----------
        file: str

        """
        self.file = file
        self._var_info = VariableSet(self._get_var_info())

        self._values = FileValueSet()
        self.encoding = detect_encod(self.file)
        self._read()

    def _read(self):

        with open(self.file, encoding=self.encoding) as f:
            section = []  # To store lines that go in the same section
            for l in f.readlines():

                if l[0] == '!':  # skip comments
                    continue

                if l[0] == '*':  # start of section

                    # Process any previously stored block
                    if section:
                        self._read_section(section)

                    # Clear the current block
                    section.clear()

                if l.strip():
                    section.append(l)  # Append current line to block

            # Read the last block too
            self._read_section(section)

    def get_var_type(self, var, sect=None):
        return self.get_var(var, sect).type_

    def get_var_lims(self, var, sect=None):
        return self.get_var(var, sect).lims

    def get_var_spc(self, var, sect=None):
        return self.get_var(var, sect).spc

    def get_var_size(self, var, sect=None):
        """
        Returns the size of a variable.

        Parameters
        ----------
        var: str
            The name of the variable.
        sect: str
            The name of the section in which this variable appears (optional; for ambiguous cases where a file
            has several variables with the same code).

        Returns
        -------
        int
            The size of the variable.

        """
        return self.get_var(var, sect).size

    def get_var_code_miss(self, var, sect=None):
        return self.get_var(var, sect).miss

    def get_var(self, var, sect=None):
        return self._var_info.get_var(var, sect)

    def get_value(self, var, sect=None, subsect=None, cond=None):
        """
        
        Parameters
        ----------
        var
        sect
        subsect
        cond

        Returns
        -------
        np.ndarray
        """
        return self._values.get_value(var, sect=sect, subsect=subsect, cond=cond)

    def get_dims_val(self, var):
        return self.get_value(var).shape

    def add_row(self, sect, subsect=None, vals=None):
        self._values.add_row(sect, subsect, vals)

    def remove_row(self, sect, subsect=None, cond=None):
        self._values.remove_row(sect, subsect, cond)

    def find_var_sect(self, var):
        return self._values.find_var_sect(var)

    def variables(self):
        return list(str(vr) for vr in self._var_info.variables())

    def to_dict(self):
        return self._values.to_dict()

    def _read_subsection(self, section_name, subblock):

        var_names = self._get_var_names(subblock[0])

        n_lines = len(subblock) - 1  # -1 for the header line (with "@" )
        lengths = [self.get_var_size(vr) for vr in var_names]
        spaces = [self.get_var_spc(vr) for vr in var_names]
        cum_lens = np.insert(np.cumsum(lengths) + np.cumsum(spaces), 0, 0)
        cutoffs = [(cum_lens[i], cum_lens[i + 1]) for i in range(len(var_names))]

        d_vals = {vr: self._gen_empty_mtrx(vr, n_lines) for vr in var_names}

        for i, l in enumerate(subblock[1:]):
            # Odd workaround necessary because several cultivar names in DSSAT are larger than the allowed space
            # and so run into the next column, which apparently isn't supposed to matter if the next column's value
            # is small enough to allow both to fit. (Really?!)
            vals = [
                (l[0 if c[0] == 0 else max(c[0], l.find(' ', c[0], c[1] - 1)):
                   None if l.find(' ', c[1] - 1) < 0 else l.find(' ', c[1] - 1)]).strip()
                for c in cutoffs]
            for vr, vl in zip(var_names, vals):
                if not vl:
                    vl = self.get_var_code_miss(vr)
                d_vals[vr][i] = vl

        l_vars = [self._var_info.get_var(vr, sect=section_name) for vr in var_names]
        l_vals = [d_vals[vr] for vr in var_names]
        subsect = ValueSubSection(l_vars, l_vals)

        self._values[section_name].add_subsection(subsect)

    def _read_section(self, section):
        section_name, section_lines = self._process_section_header(section)

        subblock = []
        for l in section_lines:  # skip first line (with "*")
            if l[0] == '@':

                if subblock:
                    self._read_subsection(section_name, subblock)
                subblock.clear()

            # Append current line to section
            if l.strip().strip('\x1a'):  # '\x1a' needed for obscure character DSSAT likes to append to .SNX/SQX
                subblock.append(l)

        if subblock:
            self._read_subsection(section_name, subblock)

    def _gen_empty_mtrx(self, var, size):
        tp = self.get_var_type(var)
        if tp == 'float':
            dtype = float
        elif tp == 'int':
            dtype = int
        elif tp == 'str' or tp == str:
            str_size = self.get_var_size(var)
            dtype = 'U{}'.format(str_size + 5)  # +5 just to be safe (with DSSAT input files you never know)
        else:
            dtype = tp

        return np.full(size, CODE_MISS, dtype=dtype)

    def _get_var_names(self, line):
        var_names = [str(vr) for vr in self._var_info]
        var_names.sort(key=len, reverse=True)

        def _strip(txt):
            return re.sub('^[|.\W]+', '', txt)

        final_names = []
        line = _strip(line[1:])  # skip initial "@"

        while line:
            try:
                name = next(vr for vr in var_names if line.startswith(vr))
            except StopIteration:
                raise ValueError(
                    'No variable matching "{line}" for file {nm}.'.format(
                        line=line[:20], nm=os.path.split(self.file)[1]
                    )
                )
            final_names.append(name)
            line = _strip(line[len(name):])

        return final_names

    def __contains__(self, item):
        return item in self._values

    def _get_var_info(self):
        """
        Return a set of variables.

        Returns
        -------
        set[Variable]
        """

        raise NotImplementedError

    @classmethod
    def matches_file(cls, file):
        """
        Checks whether a given file can be read by this class. Must be implemented in subclasses.

        Parameters
        ----------
        file: str
            The file to be read.

        Returns
        -------
        bool
            ``True`` if the file matches; ``False`` otherwise.
        """

        raise NotImplementedError

    def _process_section_header(self, lines):
        """

        Parameters
        ----------
        lines

        Returns
        -------
        tuple[str, list]
        """

        raise NotImplementedError



import os

from .file import File
from .vals import ValueSubSection
from .var import HeaderVariableSet


class InpFile(File):
    """
    Parent class for all input files, as well as for `Summary.OUT`.
    """

    ext = None  # type: str

    def __init__(self, file):
        self._header_vars = HeaderVariableSet(self._get_header_vars())
        super().__init__(file)

    def write(self, file, force=False, check=True):
        lines = []

        write = force or file != self.file or self.changed()

        if write:
            self._values.write(lines)

            with open(file, 'w', encoding=self.encoding) as f:
                f.writelines(l + "\n" for l in lines)

    def set_value(self, var, val, sect=None, subsect=None, cond=None):
        self._values.set_value(var, val, sect=sect, subsect=subsect, cond=cond)

    def changed(self):
        """
        Checks whether the file has been edited and needs to be rewritten.

        Returns
        -------
        bool
            Whether the file has been edited or not.
        """
        return self._values.changed()

    def get_var(self, var, sect=None):
        try:
            return super().get_var(var, sect=sect)
        except ValueError:
            return self._header_vars.get_var(var, sect=sect)

    def _process_section_header(self, lines):

        if lines[0][0] == '@':  # In case a section header is missing
            self._values.add_section('')
            return '', lines

        header_text = lines[0][1:].strip()  # Skip initial "*"

        match = self._header_vars.matches(header_text)
        if match:

            section_name = match.strip()
            self._values.add_section(section_name)
            header_text = header_text[len(match):]

            h_vars = self._header_vars.get_vars(match)

            l_vals = []
            for vr in h_vars:
                size = vr.size + vr.spc
                val = header_text[:size].strip()

                matr = self._gen_empty_mtrx(str(vr), size=1)
                matr[:] = val

                l_vals.append(matr)

                header_text = header_text[size:]

            header_vars_subsect = ValueSubSection(h_vars, l_vals=l_vals)
            self._values[section_name].set_header_vars(header_vars_subsect)

            return match, lines[1:]

        self._values.add_section(header_text)
        return header_text, lines[1:]

    def _get_header_vars(self):
        return {}

    @classmethod
    def matches_file(cls, file):
        """
        Checks whether a given file can be read by this class.

        Parameters
        ----------
        file: str
           The filename or full path to be read.

        Returns
        -------
        bool
           ``True`` if the file matches; ``False`` otherwise.
        """

        ext = os.path.splitext(file)[1]
        if isinstance(cls.ext, str):
            return ext.lower() == cls.ext.lower()
        return ext.lower() in [x.lower() for x in cls.ext]

    def _get_var_info(self):
        raise NotImplementedError



from tradssat.genotype.vars_._cropgro import cropgro_cul_vars, cropgro_eco_vars
from tradssat.tmpl.var import FloatVar

cul_vars_G0GRO = cropgro_cul_vars(exclude=['EXPNO', 'THRESH', 'SDPRO', 'SDLIP'])

eco_vars_G0GRO = cropgro_eco_vars()
eco_vars_G0GRO.update({
    FloatVar(
        'THRSH', 5, 1,
        info='The maximum ratio of (seed/(seed+shell)) at maturity. '
             'Causes seed to stop growing as their dry weights '
             'increase until shells are filled in a cohort. '
             '(Threshing percentage).'
    ),
    FloatVar('SDPRO', 5, 3, info='Fraction protein in seeds (g(protein)/g(seed))'),
    FloatVar('SDLIP', 5, 3, info='Fraction oil in seeds (g(oil)/g(seed))'),
})



from tradssat import SummaryOut, PlantGroOut


class DSSATResults(object):
    """
    Facilitates the reading of DSSAT results. Instead of having to read each output file individually, you can simply
    point this class to a DSSAT run output folder containing the output files and it will find the desired variables 
    for you.
    """

    def __init__(self, folder):
        """
        Initialise with the base folder.
        
        Parameters
        ----------
        folder: str
            The DSSAT run output folder.
        """

        self.folder = folder

        outfiles = [PlantGroOut]
        self._outfiles = {f: None for f in outfiles}

        self._sumoutclass = SummaryOut
        self._sumoutfile = None

    def reload(self):
        """
        Reload data (useful if a new DSSAT simulation has been run).
        """

        for out_class in self._outfiles:
            self._outfiles[out_class] = None
        self._sumoutfile = None

    def get_value(self, var, trt, t=None, at='YEAR DOY'):
        """
        Returns the value (point or time-series) of a variable from a DSSAT run.
        
        Parameters
        ----------
        var: str
            The variable name.
        trt: int
            The treatment number of interest.
        t: str | int
            The time at which one wants the results. If ``None``, results will be given for all time steps.
        at: str
            Must be one of ``DAS``, ``DOY``, or ``YEAR DOY`` (default). Only used if ``t`` is not ``None``.
        Returns
        -------
        np.ndarray:
            The value of the variable.
        """
        if t is not None:
            if at in ['DAS', 'DAP']:
                cond = {at: int(t)}
            elif at in ['DOY', 'YEAR DOY']:
                year, doy = t.split()
                cond = {'YEAR': int(year), 'DOY': int(doy)}
            else:
                raise ValueError(at)
        else:
            cond = None

        for c, f in self._outfiles.items():

            if f is None:
                try:
                    f = self._outfiles[c] = c(self.folder)
                except FileNotFoundError:
                    pass

            if var in f.variables():
                if t is None:
                    return f.get_value(var)
                return f.get_value(var, sect={'TREATMENT': trt}, cond=cond)

        raise ValueError('Output variable "{}" could not be found in any output file.'.format(var))

    def get_final_value(self, var, trt):
        """
        Returns a variable's final value. Faster than :func:`get_value` if the variable is available
        in `Summary.OUT`.

        Parameters
        ----------
        var: str
            The variable name.
        trt: int
            The treatment number of interest.

        Returns
        -------
        str | float | int
            The variable's final value.
        """

        if self._sumoutfile is None:
            self._sumoutfile = self._sumoutclass(self.folder)
        if var in self._sumoutfile.variables():
            return self._sumoutfile.get_value(var, cond={'TRNO': trt})[0]
        vals = self.get_value(var=var, trt=trt)
        return vals[-1]



import os

from tradssat import get_dssat_dir


class PeriphFileMgr(object):
    def get_value(self, var, level):
        raise NotImplementedError

    def set_value(self, var, val, level):
        raise NotImplementedError

    def variables(self):
        raise NotImplementedError


def get_dssat_subdir(subdir):
    dssat_dir = get_dssat_dir()
    if dssat_dir is not None:
        return os.path.join(dssat_dir, subdir)
    else:
        raise FileNotFoundError('No DSSAT directory available.')



import re

import numpy as np


class FileValueSet(object):
    """
    Represents the set of values in a DSSAT file.
    """

    def __init__(self):
        self._sections = {}

    def add_section(self, name):
        """
        Adds a section to the file.
        
        Parameters
        ----------
        name: str
            Name of the new section.

        """
        self._sections[name] = ValueSection(name)

    def write(self, lines):
        """
        Writes the file.
        
        Parameters
        ----------
        lines: list[str]
            List to which to write output lines.

        Returns
        -------
        list
            The modified list.
        """

        for s in self:
            s.write(lines)

        lines.append('')

        return lines

    def to_dict(self):
        """
        Converts the file to a dictionnary.
        
        Returns
        -------
        dict

        """
        return {name: sect.to_dict() for name, sect in self._sections.items()}

    def get_value(self, var, sect=None, subsect=None, cond=None):

        if isinstance(sect, str):
            return self[sect].get_value(var, subsect, cond=cond)
        else:
            if isinstance(sect, dict):
                sects = [
                    s for s in self if all(vr in s and np.all(s.get_header_var(vr) == vl) for vr, vl in sect.items())
                ]
            else:
                sects = self._sections.values()
            return next(s.get_value(var, subsect, cond=cond) for s in sects if var in s)

    def set_value(self, var, val, sect=None, subsect=None, cond=None):
        if sect is not None:
            self[sect].set_value(var, val, subsect, cond=cond)
        else:
            for s in self:
                if var in s:
                    s.set_value(var, val, subsect, cond=cond)

    def add_row(self, sect, subsect=None, vals=None):
        """
        Adds a row to the file.
        
        Parameters
        ----------
        sect: str
            Name of section.
        subsect: int
            Subsection number. If ``None``, will add row to all subsections.
        vals: dict
            Dictionnary of new row variable values.

        """
        self[sect].add_row(subsect, vals)

    def remove_row(self, sect, subsect=None, cond=None):
        self[sect].remove_row(subsect, cond)

    def find_var_sect(self, var):
        """
        Finds the section in which a variable appears.
        
        Parameters
        ----------
        var: str
            The name of the variable

        Returns
        -------
        str
            The file section name.
        """
        return next(s.name for s in self if var in s)

    def changed(self):
        """
        Detects if any variable values have been changed.
        
        Returns
        -------
        bool
        """
        return any(s.changed() for s in self)

    def __iter__(self):
        for s in self._sections.values():
            yield s

    def __contains__(self, item):
        return item in self._sections

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._sections[item]
        else:
            return next(s for s in self._sections.values() if re.match(item, s.name))


class ValueSection(object):
    """
    Represents the structure and variable values in a DSSAT file section.
    """

    def __init__(self, name):
        """

        Parameters
        ----------
        name: str
            The name of the section.
        """

        self.name = name
        self._subsections = []
        self._header_vars = HeaderValues()

    def add_subsection(self, subsect):
        """
        Add a subsection to the section.

        Parameters
        ----------
        subsect: ValueSubSection
            The new subsection.

        """
        self._subsections.append(subsect)

    def set_header_vars(self, h_vars):
        """
        Sets the section's header variables.

        Parameters
        ----------
        h_vars: ValueSubSection
            The subsection representing the header variables.

        """
        self._header_vars.set_vars(h_vars)

    def get_header_var(self, var):
        """
        Obtain the value of a header variable.

        Parameters
        ----------
        var: str
            The name of the variable

        Returns
        -------
        np.ndarray
            The value of the header variable.
        """
        return self._header_vars.get_value(var)

    def write(self, lines):
        """
        Writes the section to a list of lines.

        Parameters
        ----------
        lines: list[str]
            The list of lines to which to append this section.

        """
        lines.append(self._write_header())
        for s in self:
            s.write(lines)

        lines.append('')

    def to_dict(self):
        """
        Converts the section to a dictionnary.

        Returns
        -------
        dict
        """
        return {
            'header vars': self._header_vars.to_dict(),
            'main vars': [subsect.to_dict() for subsect in self]
        }

    def get_value(self, var, subsect=None, cond=None):

        subsect = self._valid_subsects(subsect)

        if cond is None:
            cond = {}
        req_vars = {var, *cond}

        val = []
        for s in subsect:
            sub = self[s]
            if all(vr in sub for vr in req_vars):
                filter_ = sub.filter_cond(cond)
                val.append(sub[var].val[filter_])

        return np.array(val).flatten()

    def set_value(self, var, val, subsect=None, cond=None):

        subsect = self._valid_subsects(subsect)

        if cond is None:
            cond = {}
        req_vars = {var, *cond}

        success = False
        for s in subsect:
            sub = self[s]
            if all(vr in sub for vr in req_vars):
                success = True
                sub.set_value(var, val, cond=cond)

        if not success:
            raise ValueError('Variable "{}" not found.'.format(var))

    def add_row(self, subsect=None, vals=None):
        subsect = self._valid_subsects(subsect)
        for s in subsect:
            self[s].add_row(vals)

    def remove_row(self, subsect=None, cond=None):
        subsect = self._valid_subsects(subsect)
        for s in subsect:
            self[s].remove_row(cond)

    def changed(self):
        """
        Checks whether any value has changed in the section.

        Returns
        -------
        bool
        """
        return any(s.changed() for s in self) or self._header_vars.changed()

    def _write_header(self):
        return '*' + self.name + self._header_vars.write()

    def _valid_subsects(self, subsects):
        if subsects is None:
            subsects = range(len(self._subsections))
        elif isinstance(subsects, int):
            subsects = [subsects]
        return subsects

    def __iter__(self):
        for s in self._subsections:
            yield s

    def __contains__(self, item):
        return any(item in s for s in self._subsections) or item in self._header_vars

    def __getitem__(self, item):
        return self._subsections[item]

    def __setitem__(self, key, value):
        for subsect in self:
            if key in subsect:
                subsect[key] = value


class ValueSubSection(object):
    """
    Represents the variables and values in a DSSAT file subsection.
    """

    def __init__(self, l_vars, l_vals):

        self._vars = {str(vr): VariableValue(vr, vl) for vr, vl in zip(l_vars, l_vals)}

    def set_value(self, var, val, cond=None):
        filter_ = self.filter_cond(cond)
        self._vars[str(var)].set_value(val, i=filter_)

    def add_row(self, vals=None):
        """
        Adds a row to the subsection.

        Parameters
        ----------
        vals: dict
            The values for the new row. Any missing value will be assigned the corresponding missing code for that
            variable (usually -99).
        """
        if vals is None:
            vals = {}
        for nm, vr in self._vars.items():
            val = vals[nm] if nm in vals else vr.var.miss
            vr.add_value(val)

    def remove_row(self, cond=None):
        filter_ = self.filter_cond(cond)
        for vr in self._vars.values():
            vr.remove_value(filter_)

    def filter_cond(self, cond):
        if cond is None:
            cond = {}
        return np.all(
            [self[vr] == vl for vr, vl in cond.items()], axis=0
        )

    def check_dims(self):
        """
        Checks that all variables in the subsection have the same size. (If not, the subsection cannot write to disk.)

        Raises
        -------
        ValueError
            If not all variables have the same size.
        """
        if not len(np.unique([v.size() for v in self._vars.values()])) == 1:
            raise ValueError('Not all variables in this subsection have the same size.')

    def check_vals(self):
        for vr in self:
            vr.check_val()

    def n_data(self):
        """
        Obtain the size of variables. Will fail if not all variables have the same size.

        Returns
        -------
        int
            The size of all variables in the subsection.

        Raises
        ------
        ValueError
            If not all variables have the same size.
        """

        self.check_dims()
        return self[list(self._vars)[0]].size()

    def write(self, lines):
        """
        Writes the subsection.

        Parameters
        ----------
        lines: list[str]
            The list of lines to which to append this subsection.

        """
        self.check_dims()
        self.check_vals()

        lines.append('@' + ''.join([vr.write() for vr in self]))
        for i in range(self.n_data()):
            written = [vr.write(i) for vr in self]
            for i, (x, vr) in enumerate(zip(list(written), self)):
                extra = len(x) - (vr.var.size + vr.var.spc)
                if i < len(written) - 1 and extra:
                    written[i + 1] = written[i + 1][extra:]

            line = ''.join(written)
            lines.append(line)

    def to_dict(self):
        """
        Converts the subsection to a dictionnary.

        Returns
        -------
        dict
        """
        return {nm: vr.val for nm, vr in self._vars.items()}

    def changed(self):
        """
        Checks whether any variable in the subsection has been changed.

        Returns
        -------
        bool

        """
        return any(v.changed for v in self)

    def __iter__(self):
        for vr in self._vars.values():
            yield vr

    def __contains__(self, item):
        return str(item) in self._vars

    def __getitem__(self, item):
        return self._vars[str(item)]

    def __setitem__(self, key, value):
        self.set_value(key, value)


class VariableValue(object):
    """
    Represents a DSSAT file variable.
    """

    def __init__(self, var, val):

        self.changed = False

        self.name = str(var)

        self.var = var
        self.val = val

    def set_value(self, val, i=None):
        """
        Changes the value of the variable.

        Parameters
        ----------
        val: float | int | np.ndarray
            The new value.
        i: int | np.ndarray
            Índices at which to change the value. If ``None``, all values of the variable will be changed.

        """
        if i is None:
            i = True

        if isinstance(val, np.ndarray) and (val.shape != self.val[i].shape):
            if i is not True:
                raise ValueError('Cannot set value by index when shapes do not match.')
            self.val = np.array(val)
            self.changed = True

        else:
            if np.any(val != self.val[i]):
                self.val[i] = val
                self.changed = True

    def add_value(self, val):
        """
        Adds a value to the variable's matrix.

        Parameters
        ----------
        val: int | float
            The new value.

        """
        self.val = np.append(self.val, val).astype(self.val.dtype)
        self.changed = True

    def remove_value(self, i):
        """
        Removes a value from the variable's matrix.

        Parameters
        ----------
        i: np.ndarray
            The indices of the value(s) to remove. May be a list of indices, or else a boolean mask of the same
            size as the variable.

        """
        if i.dtype == bool:
            filter_ = i
        else:
            filter_ = np.isin(np.arange(self.size()), i)
        self.val = self.val[~filter_]

    def size(self):
        """
        Returns the size of the variable.

        Returns
        -------
        int
        """
        return self.val.size

    def check_val(self):
        self.var.check_val(self.val)

    def write(self, i=None):
        """
        Converts the variable to a string.

        Parameters
        ----------
        i: int
            The index of the value to write. If ``None``, the name of the variable will be written instead.

        Returns
        -------
        str
            The properly written and formatted variable.

        """
        if i is None:
            return self.var.write()
        else:
            self.changed = False
            return self.var.write(self.val[i])

    def __eq__(self, other):
        return other == self.val


class HeaderValues(object):
    """
    Represents DSSAT file header variables and their values.
    """

    def __init__(self):
        self._subsect = None

    def set_vars(self, subsect):
        """
        Sets the variables of the header.

        Parameters
        ----------
        subsect: ValueSubSection
            The subsection with variables and their values already specified.
        """

        self._subsect = subsect

    def get_value(self, var):
        """
        Obtain the value of a header variable.

        Parameters
        ----------
        var: str
            The variable of interest.

        Returns
        -------
        np.ndarray
            The value of the variable.

        """
        return self._subsect[var].val

    def to_dict(self):
        """
        Convert to dictionnary (json) format.

        Returns
        -------
        list
            The (mostly) jsonified version of the header's variables.
        """
        if self._subsect is None:
            return {}
        else:
            return self._subsect.to_dict()

    def write(self):
        """
        Writes the header values to a string.

        Returns
        -------
        str
        """
        if self._subsect is None:
            return ''
        else:
            return ''.join([vr.write(0) for vr in self._subsect])

    def changed(self):
        """
        Checks if the header variables values have been changed.

        Returns
        -------
        bool
        """
        if self._subsect is None:
            return False
        return self._subsect.changed()

    def __contains__(self, item):
        if self._subsect is None:
            return False
        else:
            return item in self._subsect



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_MZIXM = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VRNAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar(
        'P1', 5, 1,
        info='Thermal time from seedling emergence to the end of the juvenile phase (expressed in degree days above a '
             'base temperature of 8 deg.C) during which the plant is not responsive to changes in photoperiod.'
    ),
    FloatVar(
        'P2', 5, 3,
        info='Extent to which development (expressed as days) is delayed for each hour increase in photoperiod above '
             'the longest photoperiod at which development proceeds at a maximum rate (which is considered to be 12.5 '
             'hours).'
    ),
    FloatVar(
        'P5', 5, 1,
        info='Thermal time from silking to physiological maturity (expressed in degree days above a base temperature '
             'of 8 deg.C).'
    ),
    FloatVar('G2', 5, 1, info='Maximum possible number of kernels on topmost ear.'),
    FloatVar(
        'G3', 5, 2,
        info='Kernel filling rate during the linear grain filling stage and under optimum conditions (mg/day).'
    ),
    FloatVar(
        'PHINT', 5, 2,
        info='Phylochron interval; the interval in thermal time (degree days) between successive leaf tip appearances.'
    ),
    FloatVar('AX', 5, 1, info='Leaf surface area (cm2/leaf) of largest leaf.'),
    FloatVar('ALL', 5, 1, info='Leaf longevity (degree days) of the most longevous leaf.')
}

eco_vars_MZIXM = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECONAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar('TBASE', 5, 1, info='Base temperature below which no development occurs, C'),
    FloatVar('TOPT', 5, 1, info='Temperature at which maximum development rate occurs during vegetative stages, C'),
    FloatVar('ROPT', 5, 1, info='Temperature at which maximum development rate occurs for reproductive stages, C'),
    FloatVar('P20', 5, 1, info='Daylength below which daylength does not affect development rate, hours'),
    FloatVar(
        'DJTI', 5, 1, info='Minimum days from end of juvenile stage to tassel initiation if the cultivar is not '
                           'photoperiod sensitive, days'
    ),
    FloatVar('GDDE ', 5, 1, info='Growing degree days per cm seed depth required for emergence, GDD/cm'),
    FloatVar('DSGFT', 5, 0, info='GDD from silking to effective grain filling period, C'),
    FloatVar('RUE', 5, 1, info='Radiation use efficiency, g plant dry matter/MJ PAR'),
    FloatVar('KCAN', 5, 2, info='Canopy light extinction coefficient for daily PAR.'),
    FloatVar(
        'PSTM', 5, 2, info='Parameter controlling stem C partitioning during ISTAGES 3, 4 (0-1) Higher PSTM, more DM '
                           'into stem, less into leaves'
    ),
    FloatVar(
        'PEAR', 5, 2, info='Parameter controlling ear C partitioning during ear growth  (250 GDD before flowering-end'
                           ' ISTAGE 4)'
    ),
    FloatVar('TSEN', 5, 2, info='Critical temperature below which leaf damage occurs (default 6°C)'),
    FloatVar('CDAY', 5, 2, info='Number of cold days parameter (default 15.0 )')
}



import re

from tradssat.tmpl import InpFile
from .soil_vars import main_vars, header_vars


class SoilFile(InpFile):
    """
    File reader for soil (.SOL) DSSAT files.
    """
    ext = '.SOL'

    def _get_var_info(self):
        return main_vars

    def _get_header_vars(self):
        return {re.compile('[A-Z0-9]{10}'): header_vars}



from tradssat.tmpl.var import CharacterVar, FloatVar

cul_vars_SCCAN = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ECO#', 6, info='Code for the ecotype to which this cultivar belongs (see *.eco file)'),

    FloatVar(
        'MaxPARCE', 14, 2,
        info='Maximum (no stress) radiation conversion efficiency expressed as assimilate produced before respiration, '
             'per unit PAR. (g/MJ).'
    ),
    FloatVar('APFMX', 14, 2,
             info='Maximum fraction of dry mass increments that can be allocated to aerial dry mass (t/t)'),
    FloatVar(
        'PCB', 14, 2,
        info='Partitioning coefficient: extinction coefficient of fraction of dry mass increments allocated to above '
             'ground biomass'
    ),
    FloatVar(
        'STKPFMAX', 14, 2,
        info='Fraction of daily aerial dry mass increments partitioned to stalk at high temperatures in a mature crop '
             '(t/t on a dry mass basis)'
    ),
    FloatVar(
        'DELTTMAX', 14, 2,
        info='Max. change in sucrose content per unit change in stalk mass in the unripenened section of the stalk (/t)'
    ),
    FloatVar('SUCA', 14, 2, info='Sucrose partitioning parameter: Maximum sucrose contents in the base of stalk (t/t)'),
    FloatVar(
        'TBFT', 14, 0,
        info='Sucrose partitioning:  Temperature at which partitioning of unstressed stalk mass increments to sucrose '
             'is 50% of the maximum value'
    ),
    FloatVar('Tthalfo', 14, 0, info='Thermal time to half canopy (oCd)'),
    FloatVar('TBase', 14, 0, info='Base temperature for canopy development (oCd)'),
    FloatVar(
        'LFMAX', 14, 0,
        info='Maximum number of green leaves a healthy, adequately-watered plant will have after it is old enough to '
             'lose some leaves.'
    ),
    FloatVar('MXLFAREA', 14, 0, info='Max leaf area assigned to all leaves above leaf number MXLFARNO (cm2)'),
    FloatVar('MXLFARNO', 14, 0, info='Leaf number above which leaf area is limited to MXLFAREA'),
    FloatVar('PI1', 14, 0, info='Phyllocron interval 1 (for leaf numbers below Pswitch, oC.d (base TTBASELFEX))'),
    FloatVar('PI2', 14, 0, info='Phyllocron interval 2 (for leaf numbers above Pswitch, oC.d (base TTBASELFEX))'),
    FloatVar('PSWITCH', 14, 0, info='Leaf number at which the phyllocron changes.'),
    FloatVar('MAX_POP', 14, 0, info='Maximum tiller population (stalks/m2)'),
    FloatVar('POPTT16', 14, 1, info='Stalk population at/after 1600 degree days (/m2)'),
    FloatVar('TTPLNTEM', 14, 0, info='Thermal time to emergence for a plant crop (degree C days, base TTBASEEM)'),
    FloatVar('TTRATNEM', 14, 0, info='Thermal time to emergence for a ratoon crop (degree C days, base TTBASEEM)'),
    FloatVar('CHUPIBASE', 14, 0, info='Thermal time (base TTBASEEM) from emergence to start of stalk growth'),
    FloatVar('TT_POPGROWTH', 14, 1, info='Thermal time to peak tiller population (deg C days, TTBASEPOP)'),
    FloatVar(
        'LG_AMBASE', 14, 0,
        info='Aerial mass (fresh mass of stalks, leaves, and water attached to them) at which lodging starts; t/ha'
    ),
}

eco_vars_SCCAN = {
    CharacterVar('ECO#', 6, spc=0, info='Code for the ecotype to which a cultivar belongs (see *.cul file)'),
    CharacterVar('ECO-NAME', 17, header_fill='.', info='Name of the ecotype, which is referenced from *.CUL file'),

    FloatVar(
        'DELTTMAX', 14, 2,
        info='Max. change in sucrose content per unit change in stalk mass in the unripenened section'
    ),
    FloatVar('SWDF2AMP', 14, 1, info='Sucrose partitioning sensitivity to water stress parameter'),
    FloatVar('CS_CNREDUC', 14, 1, info='Canopy reduction due to water stress'),
    FloatVar('CS_CNPERIOD', 14, 0, info='Canopy water stress period (days)'),
    FloatVar('Tthalfa', 14, 0, info='Half canopy thermal time adjustment for row width'),
    FloatVar('dPERdT', 14, 3, info='Change in plant extension rate (mm/h) per unit change in temperature (oC)'),
    FloatVar('EXTCFN', 14, 2, info='Maximum canopy light extinction coefficient'),
    FloatVar('EXTCFST', 14, 2, info='Minimum canopy light extinction coefficient'),
    FloatVar(
        'LFNMXEXT', 14, 0,
        info='Leaf number (including dead leaves still attached) at which maximum light extinction occurs'
    ),
    FloatVar('AREAMX_CF(1)', 14, 0, info='Cultivar parameter for quadratic equation defining maximum leaf area'),
    FloatVar('AREAMX_CF(2)', 14, 1, info='Cultivar parameter for quadratic equation defining maximum leaf area'),
    FloatVar('AREAMX_CF(3)', 14, 1, info='Cultivar parameter for quadratic equation defining maximum leaf area'),
    FloatVar('WIDCOR', 14, 0, info='Parameter affecting the width of leaves'),
    FloatVar('WMAX_CF(1)', 14, 4, info='Cultivar parameter for quadratic equation defining leaf width'),
    FloatVar('WMAX_CF(2)', 14, 3, info='Cultivar parameter for quadratic equation defining leaf width'),
    FloatVar('WMAX_CF(3)', 14, 2, info='Cultivar parameter for quadratic equation defining leaf width'),
    FloatVar(
        'POPCF(1)', 14, 3,
        info='Stalk population coefficient, in ideal conditions (no stresses), as function of th.time'
    ),
    FloatVar(
        'POPCF(2)', 14, 5,
        info='Stalk population coefficient, in ideal conditions (no stresses), as function of th.time'
    ),
    FloatVar(
        'POPDECAY', 14, 3,
        info='Tiller senescence rate expressed as the fraction of tillers above the future mature tiller population '
             '(at a thermal time of 1600 oC Adjust by comparing simulated and observed tiller population. Varies from '
             '3.0 to 5.0'
    ),
    FloatVar('TTBASEEM', 14, 1, info='Base temperature for emergence (oC)'),
    FloatVar('TTBASELFEX', 14, 1, info='Base temperature for leaf phenology (oC)'),
    FloatVar('TTBASEPOP', 14, 1, info='Base temperature for stalk phenology (oC)'),
    FloatVar('TBASEPER', 14, 3, info='Base temperature for plant extension (oC)'),
    FloatVar('LG_AMRANGE', 14, 0, info='Range in aerial mass from the start to the end of lodging (t/ha)'),
    FloatVar('LG_GP_REDUC', 14, 2,
             info='Reduction in gross photosynthesis due to full lodging, as a fraction (Singh, et al.)'),
    FloatVar('LDG_FI_REDUC', 14, 1, info='Reduction in fractional interception by the canopy due to full lodging'),
    FloatVar('LMAX_CF(1)', 14, 3),
    FloatVar('LMAX_CF(2)', 14, 1),
    FloatVar('LMAX_CF(3)', 14, 1),
    FloatVar('MAXLFLENGTH', 14, 0),
    FloatVar('MAXLFWIDTH', 14, 1)
}



from tradssat.tmpl.var import CharacterVar

cul_vars_RIORZ = {
    CharacterVar('VAR#', 6, spc=0, info='Identification code or number for the specific cultivar.'),
    CharacterVar('VAR-NAME', 16, header_fill='.', info='Name of cultivar.'),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used for calibration.'),
    CharacterVar('ORYZA file name', 15, info='Name of ORYZA2000 crop file'),
}



from tradssat.tmpl.var import CharacterVar, FloatVar, IntegerVar

cul_vars_SCCSP = {  # Todo: add info for all vars
    CharacterVar(
        'VAR#', 6, spc=0, info='Identification code or number for a specific cultivar used by  DSSAT program.'
    ),
    CharacterVar(
        'VAR-NAME', 16, header_fill='.', info='Name of cultivar as recognized commercially or within industry.'
    ),
    CharacterVar('EXPNO', 5, miss='.', info='Number of experiments used to estimate cultivar parameters.'),
    CharacterVar('ECO#', 6, info='Ecotype code of this cultivar in the ECO input file'),

    FloatVar(
        'LFMAX', 5, 3,
        info='Maximum leaf photosynthesis rate, 30 C, 350 ppm CO2 and high '
             'light (used in leaf-level ET routine - Not yet in CASUPRO). mg CO2/m2-s'),
    FloatVar(
        'PHTMX', 5, 1, info='Maximum amount of CH20 which can be produced if PAR is very high, g[CH2O]/m2-d'
    ),
    FloatVar('Stalk', 5, 3),
    FloatVar('Sucro', 5, 3),
    IntegerVar('Null1', 5),
    FloatVar('PLF1', 5, 3),
    FloatVar('PLF2', 5, 3),
    FloatVar('Gamma', 5, 3),
    FloatVar('StkB', 5, 3),
    FloatVar('StkM', 5, 3),
    IntegerVar('Null3', 5),
    FloatVar('SIZLF', 5, 2),
    FloatVar('PLF2', 5, 3),
    FloatVar('LIsun', 5, 3),
    FloatVar('LIshd', 5, 3),
    IntegerVar('Null4', 5),
    FloatVar('TB(1)', 5, 3),
    FloatVar('TO1(1)', 5, 2),
    FloatVar('TO2(1)', 5, 2),
    FloatVar('TM(1)', 5, 2),
    FloatVar('PI1', 5, 2),
    FloatVar('PI2', 5, 2),
    FloatVar('DTPI', 5, 1),
    FloatVar('LSFAC', 5, 3),
    IntegerVar('Null5', 5),
    FloatVar('LI1', 5, 3),
    FloatVar('TELOM', 5, 1),
    FloatVar('TB(2)', 5, 2),
    FloatVar('TO1(2)', 5, 2),
    FloatVar('TO2(2)', 5, 2),
    FloatVar('TM(2)', 5, 2),
    FloatVar('Ph1P', 5, 1),
    FloatVar('Ph1R', 5, 1),
    FloatVar('Ph2', 5, 3),
    FloatVar('Ph3', 5, 0),
    FloatVar('Ph4', 5, 0),
    FloatVar('StHrv', 5, 3),
    FloatVar('RTNFAC', 5, 3),
    FloatVar('MinGr', 5, 2),
    IntegerVar('Null7', 5),
    FloatVar('RES30C', 5, 3),
    FloatVar('RLF30C', 5, 3),
    FloatVar('R30C2', 5, 3),

}



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class PlantGroOut(OutFile):
    """
    File reader for DSSAT plant growth (PLANTGRO.OUT) output files.
    """
    filename = 'PlantGro.OUT'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('TREATMENT', 2, info='Treatment number'),

    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 5, info='Day after start'),
    IntegerVar('DAP', 5, info='Days after planting'),
    FloatVar('L#SD', 6, 0, info='Leaf number'),
    IntegerVar('GSTD', 6, info='Growth stage'),
    FloatVar('LAID', 6, 2, info='Leaf Area Index'),
    IntegerVar('CWAD', 6, info='Tops dry weight, kg/Ha'),
    IntegerVar('VWAD', 6, info='Veg dry weight, kg/Ha'),
    IntegerVar('LWAD', 6, info='Leaf dry weight, kg/Ha'),
    IntegerVar('SWAD', 6, info='Stem dry weight, kg/Ha'),
    IntegerVar('FLWAD', 6, info='Flower dry weight, kg/Ha'),
    IntegerVar('FWAD', 6, info='Fruit dry weight, kg/Ha'),
    IntegerVar('CRAD', 6, info='Crown dry weight, kg/Ha'),
    IntegerVar('BWAD', 6, info='Basal dry weight, kg/Ha'),
    IntegerVar('SUGD', 6, info='Suck dry weight, kg/Ha'),
    IntegerVar('RWAD', 6, info='Root dry weight, kg/Ha'),
    FloatVar('HIAD', 6, 3, info='Harvest index'),
    IntegerVar('EYWAD', 6, info='Eye Weight, kg/Ha'),
    IntegerVar('EY#AD', 6, info='Eye number'),
    FloatVar('WSPD', 6, 3, info='Water stress in photosynthesis'),
    FloatVar('WSGD', 6, 3, info='Water stress in growth'),
    FloatVar('NSTD', 6, 3, info='Nitrogen stress'),
    FloatVar('LN%D', 6, 2, info='Leaf Nitrogen percentage'),
    FloatVar('SLAD', 6, 1, info='Specific Leaf area'),
    FloatVar('RDPD', 6, 1, info='Root depth, m'),
    FloatVar('RL1D', 6, 2, info='Level 1 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL2D', 6, 2, info='Level 2 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL3D', 6, 2, info='Level 3 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL4D', 6, 2, info='Level 4 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL5D', 6, 2, info='Level 5 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL6D', 6, 2, info='Level 6 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL7D', 6, 2, info='Level 7 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL8D', 6, 2, info='Level 8 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL9D', 6, 2, info='Level 9 Root Length density, cm3/cm3 of soil'),
    FloatVar('RL10', 6, 2, info='Level 10 Root Length density, cm3/cm3 of soil')

}



from tradssat.genotype.vars_._cropgro import cropgro_cul_vars, cropgro_eco_vars
from tradssat.tmpl.var import CharacterVar

cul_vars_COGRO = cropgro_cul_vars()

eco_vars_COGRO = cropgro_eco_vars()
eco_vars_COGRO.add(CharacterVar('KCAN', 5, info='Who knows'))



from tradssat.tmpl import InpFile
from .mth_vars import vars_


class MTHFile(InpFile):
    """
    Reader for DSSAT monthly weather (.MTH) files.
    """

    ext = '.MTH'

    def _get_var_info(self):
        return vars_



import re

from tradssat.tmpl.var import CharacterVar, FloatVar, IntegerVar

TRT_HEAD = re.compile('TREATMENTS(\W+[-]+FACTOR LEVELS[-]+)?')
GENERAL = 'GENERAL'

header_vars = [
    CharacterVar('EXPCODE', 10, info='Experiment identifier'),
    CharacterVar('ENAME', 60, info='Experiment name')
]

main_vars = {

    # General
    CharacterVar('PEOPLE', 75, sect=GENERAL, info='Names of scientists'),
    CharacterVar('ADDRESS', 75, sect=GENERAL, info='Contact address of principal scientist'),
    CharacterVar('SITE', 75, sect=GENERAL, info='Name and location of experimental site(s)'),

    FloatVar('PAREA', 6, 1, sect=GENERAL, spc=3, info='Gross plot area per rep, m-2'),
    IntegerVar('PRNO', 5, sect=GENERAL, info='Rows per plot'),
    FloatVar('PLEN', 5, 1, sect=GENERAL, info='Plot length, m'),
    IntegerVar('PLDR', 5, sect=GENERAL, info='Plots relative to drains, degrees'),
    IntegerVar('PLSP', 5, sect=GENERAL, info='Plot spacing, cm'),
    CharacterVar('PLAY', 5, sect=GENERAL, info='Plot layout'),
    FloatVar('HAREA', 5, 1, sect=GENERAL, info='Harvest area, m-2'),
    IntegerVar('HRNO', 5, sect=GENERAL, info='Harvest row number'),
    FloatVar('HLEN', 5, 1, sect=GENERAL, info='Harvest row length, m'),
    CharacterVar('HARM', 15, sect=GENERAL, info='Harvest method'),

    CharacterVar('NOTES', 500, sect=GENERAL, info='Notes'),

    # Treatments
    IntegerVar('N', 2, spc=0, sect=TRT_HEAD, info='Treatment number'),
    IntegerVar('R', 1, sect=TRT_HEAD, info='Rotation component: number (default=1)'),
    IntegerVar('O', 1, sect=TRT_HEAD, info='Rotation component: option (default=1)'),
    IntegerVar('C', 1, sect=TRT_HEAD, info='Crop component number (default = 0)'),
    CharacterVar('TNAME', 25, header_fill='.', info='Treatment name'),
    IntegerVar('CU', 2, info='Cultivar level'),
    IntegerVar('FL', 2, info='Field level'),
    IntegerVar('SA', 2, info='Soil analysis level'),
    IntegerVar('IC', 2, info='Initial conditions level'),
    IntegerVar('MP', 2, info='Planting level'),
    IntegerVar('MI', 2, info='Irrigation level'),
    IntegerVar('MF', 2, info='Fertilizer level'),
    IntegerVar('MR', 2, info='Residue level'),
    IntegerVar('MC', 2, info='Chemical applications level'),
    IntegerVar('MT', 2, info='Tillage and rotations level'),
    IntegerVar('ME', 2, info='Environmental modifications level'),
    IntegerVar('MH', 2, info='Harvest level'),
    IntegerVar('SM', 2, info='Simulation control level'),

    # Cultivars
    IntegerVar('C', 2, spc=0, sect='CULTIVARS', info='Cultivar level'),
    CharacterVar('CR', 2, info='Crop code'),
    CharacterVar('INGENO', 6, info='Cultivar identifier (Institute code + Number)'),
    CharacterVar('CNAME', 16, info='Cultivar name'),

    # Fields
    IntegerVar('L', 2, spc=0, sect='FIELDS', info='Field level '),
    CharacterVar('ID_FIELD', 8, info='Field ID (Institute + Site + Field)'),
    CharacterVar('WSTA', 8, info='Weather station code (Institute+Site)'),
    CharacterVar('FLSA', 5, info='Slope and aspect, degrees from horizontal plus direction (W, NW, etc.)'),
    FloatVar('FLOB', 5, 0, info='Obstruction to sun, degrees'),
    CharacterVar('FLDT', 5, info='Drainage type, code'),
    FloatVar('FLDD', 5, 0, info='Drain depth, cm'),
    FloatVar('FLDS', 5, 0, info='Drain spacing, m'),
    CharacterVar('FLST', 5, info='Surface stones(Abundance,%+Size,S,M,L)'),
    CharacterVar('SLTX', 5, info='Soil texture'),
    FloatVar('SLDP', 5, 0, info='Soil depth, cm'),
    CharacterVar('ID_SOIL', 10, info='Soil ID (Institute+Site+Year+Soil)'),
    CharacterVar('FLNAME', 25, info='Field level name'),  # size is guess

    # Todo: define and verify
    FloatVar('XCRD', 15, 5, header_fill='.'),
    FloatVar('YCRD', 15, 5, header_fill='.'),
    FloatVar('ELEV', 9, 2, header_fill='.'),
    FloatVar('AREA', 17, 1, header_fill='.'),
    IntegerVar('SLEN', 5, header_fill='.'),
    FloatVar('FLWR', 5, 1, header_fill='.'),
    FloatVar('SLAS', 5, 1, header_fill='.'),
    CharacterVar('FLHST', 5),
    CharacterVar('FHDUR', 5),

    # Soil analysis
    IntegerVar('A', 2, spc=0, sect='SOIL ANALYSIS', info='Soil analysis level'),
    IntegerVar('SADAT', 5, info='Analysis date, year + days from Jan. 1'),
    CharacterVar('SMHB', 5, info='pH in buffer determination method, code'),
    CharacterVar('SMPX', 5, info='Phosphorus determination method, code'),
    CharacterVar('SMKE', 5, info='Potassium determination method, code'),
    FloatVar('SABL', 5, 0, info='Depth, base of layer, cm'),
    FloatVar('SADM', 5, 1, info='Bulk density, moist, g cm-3'),
    FloatVar('SAOC', 5, 2, info='Organic carbon, g kg-1 '),
    FloatVar('SANI', 5, 2, info='Total nitrogen, g kg-1'),
    FloatVar('SAPHW', 5, 1, info='pH in water'),
    FloatVar('SAPHB', 5, 1, info='pH in buffer'),
    FloatVar('SAPX', 5, 1, info='Phosphorus, extractable, mg kg-1 '),
    FloatVar('SAKE', 5, 1, info='Potassium, exchangeable, cmol kg-1'),
    FloatVar('SASC', 5, 1),
    CharacterVar('SANAME', 25, info='Soil analysis level name'),  # size is guess

    # Initial conditions
    IntegerVar('C', 2, spc=0, sect='INITIAL CONDITIONS', info='Initial conditions level'),
    CharacterVar('PCR', 5, info='Previous crop code'),
    IntegerVar('ICDAT', 5, info='Initial conditions measurement date, year + days'),
    FloatVar('ICRT', 5, 0, info='Root weight from previous crop, kg ha-1'),
    FloatVar('ICND', 5, 0, info='Nodule weight from previous crop, kg ha-1'),
    FloatVar('ICRN', 5, 2, info='Rhizobia number, 0 to 1 scale (default = 1)'),
    FloatVar('ICRE', 5, 2, info='Rhizobia effectiveness, 0 to 1 scale (default = 1)'),

    # todo: define and verify
    FloatVar('ICWD', 5, 2),
    FloatVar('ICRES', 5, 0),
    FloatVar('ICREN', 5, 2),
    FloatVar('ICREP', 5, 0),
    FloatVar('ICRIP', 5, 0),
    FloatVar('ICRID', 5, 0),
    CharacterVar('ICNAME', 25, info='Inicial condition level name'),

    FloatVar('ICBL', 5, 0, info='Depth, base of layer, cm'),
    FloatVar('SH2O', 5, 3, info='Water, cm3 cm-3 x 100 volume percent'),
    FloatVar('SNH4', 5, 1, info='Ammonium, KCl, g elemental N Mg-1 soil'),
    FloatVar('SNO3', 5, 1, info='Nitrate, KCl, g elemental N Mg-1 soil'),

    # Planting details
    IntegerVar('P', 2, spc=0, sect='PLANTING DETAILS', info='Planting level number'),
    IntegerVar('PDATE', 5, info='Planting date, year + days from Jan. 1'),
    IntegerVar('EDATE', 5, info='Emergence date, earliest treatment'),
    FloatVar('PPOP', 5, 1, info='Plant population at seeding, plants m-2'),
    FloatVar('PPOE', 5, 1, info='Plant population at emergence, plants m-2 '),
    CharacterVar(
        'PLME', 1, spc=5, info='Planting method, transplant (T), seed (S), pregerminated seed (P) or nursery (N)'
    ),
    CharacterVar('PLDS', 1, spc=5, info='Planting distribution, row (R), broadcast (B) or hill (H) '),
    FloatVar('PLRS', 5, 0, info='Row spacing, cm'),
    FloatVar('PLRD', 5, 0, info='Row direction, degrees from N'),
    FloatVar('PLDP', 5, 1, info='Planting depth, cm'),
    FloatVar('PLWT', 5, 0, info='Planting material dry weight, kg ha-1'),
    FloatVar('PAGE', 5, 0, info='Transplant age, days '),
    FloatVar('PENV', 5, 1, info='Temp. of transplant environment, °C'),
    FloatVar('PLPH', 5, 1, info='Plants per hill (if appropriate)'),

    # Todo: check and document
    FloatVar('SPRL', 5, 0),
    CharacterVar('PLNAME', 35, info='Planting details level name'),

    # IRRIGATION AND WATER MANAGEMENT
    IntegerVar('I', 2, spc=0, sect='IRRIGATION AND WATER MANAGEMENT', info='Irrigation level'),
    FloatVar('EFIR', 5, 1, info='Irrigation application efficiency, fraction'),
    FloatVar('IDEP', 5, 0, info='Management depth for automatic application, cm'),
    FloatVar('ITHR', 5, 0, info='Threshold for automatic appl., % of max. available'),
    FloatVar('IEPT', 5, 0, info='End point for automatic appl., % of max. available'),
    CharacterVar('IOFF', 5, info='End of applications, growth stage '),
    CharacterVar('IAME', 5, info='Method for automatic applications, code'),
    FloatVar('IAMT', 5, 0, info='Amount per irrigation if fixed, mm'),

    # Todo: check size
    CharacterVar('IRNAME', 25, info='Irrigation level name'),

    IntegerVar('IDATE', 5, info='Irrigation date, year + day or days from planting'),
    CharacterVar('IROP', 5, info='Irrigation operation, code'),
    FloatVar(
        'IRVAL', 5, 0,
        info='Irrigation amount, depth of water/water table, bund height, or percolation rate, mm or mm day -1'
    ),

    # FERTILIZERS (INORGANIC)
    IntegerVar('F', 2, spc=0, sect='FERTILIZERS (INORGANIC)', info='Fertilizer application level'),
    IntegerVar('FDATE', 5, info='Fertilization date, year + day or days from planting'),
    CharacterVar('FMCD', 5, info='Fertilizer material, code'),
    CharacterVar('FACD', 5, info='Fertilizer application/placement, code'),
    FloatVar('FDEP', 5, 0, info='Fertilizer incorporation/application depth, cm'),
    FloatVar('FAMN', 5, 0, info='N in applied fertilizer, kg ha-1'),
    FloatVar('FAMP', 5, 0, info='P in applied fertilizer, kg ha-1'),
    FloatVar('FAMK', 5, 0, info='K in applied fertilizer, kg ha-1'),
    FloatVar('FAMC', 5, 0, info='Ca in applied fertilizer, kg ha-1'),
    FloatVar('FAMO', 5, 0, info='Other elements in applied fertilizer, kg ha-1'),
    CharacterVar('FOCD', 5, info='Other element code, e.g., MG'),

    # Todo: check size
    CharacterVar('FERNAME', 25, info='Fertilizer level name'),

    # RESIDUES AND ORGANIC FERTILIZER
    IntegerVar('R', 2, spc=0, sect='RESIDUES AND ORGANIC FERTILIZER', info='Residue management level'),
    IntegerVar('RDATE', 5, info='Incorporation date, year + days'),
    CharacterVar('RCOD', 5, info='Residue material, code'),
    FloatVar('RAMT', 5, 0, info='Residue amount, kg ha-1'),
    FloatVar('RESN', 5, 2, info='Residue nitrogen concentration, %'),
    FloatVar('RESP', 5, 2, info='Residue phosphorus concentration, %'),
    FloatVar('RESK', 5, 2, info='Residue potassium concentration, %'),
    FloatVar('RINP', 5, 0, info='Residue incorporation percentage, %'),
    FloatVar('RDEP', 5, 0, info='Residue incorporation depth, cm'),

    # Todo: check
    FloatVar('RMET', 5, 0, info=''),
    CharacterVar('RENAME', 25, info='Residue management level name'),

    # CHEMICAL APPLICATIONS
    IntegerVar('C', 2, spc=0, sect='CHEMICAL APPLICATIONS', info='Chemical applications level'),
    IntegerVar('CDATE', 5, info='Application date, year + day or days from planting'),
    CharacterVar('CHCOD', 5, info='Chemical material, code'),
    FloatVar('CHAMT', 5, 2, info='Chemical application amount, kg ha-1'),
    CharacterVar('CHME', 5, info='Chemical application method, code'),
    FloatVar('CHDEP', 5, 0, info='Chemical application depth, cm'),
    CharacterVar('CHT', 5, info='Chemical targets'),
    CharacterVar('CHNAME', 25, info='Chemical application level name'),  # todo: check size

    # TILLAGE
    IntegerVar('T', 2, spc=0, info='Tillage level'),
    IntegerVar('TDATE', 5, info='Tillage date, year + day'),
    CharacterVar('TIMPL', 5, info='Tillage implement, code'),
    FloatVar('TDEP', 5, 0, info='Tillage implement, code'),

    # ENVIRONMENT MODIFICATIONS
    IntegerVar('E', 2, spc=0, info='Environment modifications level '),
    IntegerVar('ODATE', 5, info='Modification date, year + day or days from planting'),
    # The following are officially 2 variables each in DSSAT (code + value). However, as this is almost impossible
    # to implement sanely here (no space between variable names, and repeated names in the same section, we will treat
    # these as character values including both the code and value.
    CharacterVar('EDAY', size=5, info='Daylength adjustment, factor (A, S, M, R) + h'),
    CharacterVar('ERAD', size=5, info='Radiation adjustment, factor (A, S, M, R) + MJ m-2 d-1'),
    CharacterVar('EMAX', size=5, info='Temperature (maximum) adjustment, factor (A, S, M, R) + °C'),
    CharacterVar('EMIN', size=5, info='Temperature (minimum) adjustment, factor (A, S, M, R) + °C'),
    CharacterVar('ERAIN', size=5, info='Precipitation adjustment, factor (A, S, M, R) + mm'),
    CharacterVar('ECO2', size=5, info='CO2 adjustment, factor (A, S, M, R) + vpm'),
    CharacterVar('EDEW', size=5, info='Humidity (dew pt) adjustment, factor (A, S, M, R) + °C'),
    CharacterVar('EWIND', size=5, info='Daylength adjustment, factor (A, S, M, R) + h'),
    CharacterVar('ENVNAME', size=25, info='Environmental modification level name'),  # todo: check size

    # HARVEST DETAILS
    IntegerVar('H', 2, spc=0, info='Harvest level'),
    IntegerVar('HDATE', 5, info='Harvest date, year + day or days from planting'),
    CharacterVar('HSTG', 5, info='Harvest stage'),
    CharacterVar('HCOM', 5, info='Harvest component, code'),
    CharacterVar('HSIZE', 5, info='Harvest size group, code'),
    FloatVar('HPC', 5, 0, info='Harvest percentage, %'),

    # Todo: check
    FloatVar('HBPC', 5, 0, info=''),
    CharacterVar('HNAME', 25, info='Harvest details level name'),

    # SIMULATION CONTROLS
    IntegerVar('N', 2, spc=0, sect='SIMULATION CONTROLS', info='Simulation control level number'),
    CharacterVar('GENERAL', 11, info='Identifier'),
    IntegerVar('NYERS', 2, spc=4, info='Years'),
    IntegerVar('NREPS', 2, spc=4, info='Replications'),
    CharacterVar(
        'START', 1, spc=5,
        info='Start of Simulation, code: '
             'E = On reported emergence date; '
             'I = When initial conditions measured; '
             'P = On reported planting date; '
             'S = On specified date'
    ),
    IntegerVar('SDATE', 5, info='Date, year + day (if needed)'),
    IntegerVar('RSEED', 5, info='Random number seed'),
    CharacterVar('SNAME', 25, header_fill='.', info='Title'),
    CharacterVar('SMODEL', 5, info=''),  # todo: check

    CharacterVar('OPTIONS', 11, info='Identifier'),
    CharacterVar('WATER', 1, spc=5, info='Water (Y = yes; N = no)'),
    CharacterVar('NITRO', 1, spc=5, info='Nitrogen (Y = yes; N = no)'),
    CharacterVar('SYMBI', 1, spc=5, info='Symbiosis (Y= yes, N= no, U= unlimited N)'),
    CharacterVar('PHOSP', 1, spc=5, info='Phosphorus (Y = yes; N = no)'),
    CharacterVar('POTAS', 1, spc=5, info='Potassium (Y = yes; N = no)'),
    CharacterVar('DISES', 1, spc=5, info='Diseases and other pests (Y = yes; N = no)'),
    CharacterVar('CHEM', 1, spc=5, info='Chemical applications (Y = yes; N = no)'),
    CharacterVar('TILL', 1, spc=5, info='Tillage (Y = yes; N = no)'),
    CharacterVar('CO2', 1, spc=5, miss='', info='CO2 effects (Y = yes; N = no)'),

    CharacterVar('METHODS', 11, info='Identifier'),
    CharacterVar(
        'WTHER', 1, spc=5,
        info='Weather: '
             'M = Measured data, as recorded; '
             'G = Simulated data, stored as *.WTG files; '
             'S = Simulated data (Internal weather generator using monthly inputs); '
             'W = Simulated data (Internal WGEN weather generator) '
    ),
    CharacterVar(
        'INCON', 1, spc=5,
        info='Initial Soil Conditions: '
             'M = As reported; '
             'S = Simulated outputs from previous model run'
    ),
    CharacterVar(
        'LIGHT', 1, spc=5,
        info='Light interception: '
             'E = Exponential with LAI; '
             "H = ‘Hedgerow’ calculations"
    ),
    CharacterVar(
        'EVAPO', 1, spc=5,
        info='Evaporation: '
             'P = FAO - Penman; '
             'R = Ritchie modification of Priestley-Taylor'
    ),
    CharacterVar(
        'INFIL', 1, spc=5,
        info='Infiltration: '
             'R = Ritchie method; '
             'S = Soil Conservation Service routines'
    ),
    CharacterVar(
        'PHOTO', 1, spc=5,
        info='Photosynthesis: '
             'C = Canopy photosynthesis response curve; '
             'R = Radiation use efficiency; '
             'L = Leaf photosynthesis response curve'
    ),

    # Todo: check and document
    CharacterVar(
        'HYDRO', 1, spc=5,
        info=''
    ),
    IntegerVar('NSWIT', 1, spc=5, info=''),
    CharacterVar(
        'MESOM', 1, spc=5, info=''
    ),
    CharacterVar(
        'MESEV', 1, spc=5, info=''
    ),
    IntegerVar('MESOL', 1, spc=5),

    CharacterVar('MANAGEMENT', 11, info='Identifier'),
    CharacterVar(
        'PLANT', 1, spc=5,
        info='Planting/Transplanting: '
             'A = Automatic when conditions satisfactory; '
             'R = On reported date'
    ),
    CharacterVar(
        'IRRIG', 1, spc=5,
        info='Irrigation and Water Management: '
             'A = Automatic when required; '
             'N = Not irrigated; '
             'F = Automatic with fixed amounts at each irrigation date; '
             'R = On reported dates; '
             'D = As reported, in days after planting'
    ),
    CharacterVar(
        'FERTI', 1, spc=5,
        info='Fertilization: '
             'A = Automatic when required; '
             'N = Not fertilized; '
             'F = Automatic with fixed amounts at each fertilization date; '
             'R = On reported dates; '
             'D = As reported, in days after planting'
    ),
    CharacterVar(
        'RESID', 1, spc=5,
        info='Residue applications: '
             'A = Automatic for multiple years/crop sequences; '
             'N = No applications; '
             'F = Automatic with fixed amounts at each residue application date; '
             'R = On reported dates; '
             'D = As reported, in days after planting'
    ),
    CharacterVar(
        'HARVS', 1, spc=5,
        info='Harvest: '
             'A = Automatic when conditions satisfactory; '
             'G = At reported growth stage(s); '
             'M = At maturity; '
             'R = On reported date(s); '
             'D = On reported days after planting'
    ),

    CharacterVar('OUTPUTS', 11, info='Identifier'),
    CharacterVar('FNAME', 1, spc=5, info='Experiment (Y = yes, files named with the experiment code; N = no)'),
    CharacterVar('OVVEW', 1, spc=5, info='Overview (Y = yes, new; A = append; N = no)'),
    CharacterVar('SUMRY', 1, spc=5, info='Summary (Y = yes, new; A = append; N = no)'),
    IntegerVar('FROPT', 2, spc=4, info='Frequency of output (days)'),
    CharacterVar('GROUT', 1, spc=5, info='Growth (Y = yes; N = no)'),
    CharacterVar('CAOUT', 1, spc=5, info='Carbon (Y = yes; N = no)'),
    CharacterVar('WAOUT', 1, spc=5, info='Water (Y = yes; N = no)'),
    CharacterVar('NIOUT', 1, spc=5, info='Nitrogen (Y = yes; N = no)'),
    CharacterVar('MIOUT', 1, spc=5, info='Phosphorous (Y = yes; N = no)'),
    CharacterVar('DIOUT', 1, spc=5, info='Diseases and other pests (Y = yes; N = no)'),

    # todo: check and document
    CharacterVar('VBOSE', 1, spc=5, info='Wide (Y) or 80-column (N) daily outputs'),
    CharacterVar('CHOUT', 1, spc=5),
    CharacterVar('OPOUT', 1, spc=5),
    CharacterVar('FMOPT', 1, spc=5),

    # Automatic Management
    CharacterVar('AUTOMATIC MANAGEMENT', 0),  # No idea why this variable even exists

    CharacterVar('PLANTING', 11, info='Identifier'),
    IntegerVar('PFRST', 5, info='Earliest, year and day of year (YRDOY)'),
    IntegerVar('PLAST', 5, info='Latest, year and day of year (YRDOY)'),
    FloatVar('PH2OL', 5, 0, info='Lowermost soil water, % '),
    FloatVar('PH2OU', 5, 0, info='Uppermost soil water, %'),
    FloatVar('PH2OD', 5, 0, info='Management depth for water, cm'),
    FloatVar('PSTMX', 5, 0, info='Max. soil temp. (10 cm av.), °C'),
    FloatVar('PSTMN', 5, 0, info='Min. soil temp. (10 cm av.), °C'),

    CharacterVar('IRRIGATION', 11, info='Identifier'),
    FloatVar('IMDEP', 5, 0, info='Management depth, cm'),
    FloatVar('ITHRL', 5, 0, info='Threshold, % of maximum available'),
    FloatVar('ITHRU', 5, 0, info='End point, % of maximum available'),
    CharacterVar('IROFF', 5, info='End of applications, growth stage'),
    CharacterVar('IMETH', 5, info='Method, code'),
    FloatVar('IRAMT', 5, 0, info='Amount per irrigation, if fixed, mm'),
    FloatVar('IREFF', 5, 2, info='Irrigation application efficiency, fraction'),

    CharacterVar('NITROGEN', 11, info='Identifier'),
    FloatVar('NMDEP', 5, 0, info='Application depth, cm'),
    FloatVar('NMTHR', 5, 0, info='Threshold, N stress factor, %'),
    FloatVar('NAMNT', 5, 0, info='Amount per application, kg N ha-1'),
    CharacterVar('NCODE', 5, info='Material, code'),
    CharacterVar('NAOFF', 5, info='End of applications, growth stage'),

    CharacterVar('RESIDUES', 11, info='Identifier'),
    FloatVar('RIPCN', 5, 0, info='Incorporation percentage, % of remaining'),
    IntegerVar('RTIME', 5, info='Incorporation time, days after harvest'),
    FloatVar('RIDEP', 5, 0, info='Incorporation depth, cm'),

    CharacterVar('HARVEST', 11, info='Identifier'),
    IntegerVar('HFRST', 5, info='Earliest, days after maturity'),
    IntegerVar('HLAST', 5, info='Latest, year and day of year (YRDOY)'),
    FloatVar('HPCNP', 5, 0, info='Percentage of product harvested, %'),
    FloatVar('HPCNR', 5, 0, info='Percentage of residue harvested, %')

}



from setuptools import setup, find_packages


def read_txt(f):
    with open(f, 'r', encoding='UTF-8') as d:
        return d.read()


setup(
    name='tradssat',
    version=read_txt('tradssat/version.txt'),
    packages=find_packages(),
    url='',
    license='MIT',
    author=read_txt('tradssat/AUTHORS.txt'),
    author_email='julien.malard@mail.mcgill.ca',
    description='DSSAT input and output file reader and writer',
    long_description=read_txt('README.md'),
    setup_requires=[],
    install_requires=['numpy', 'chardet'],
    package_data={'tradssat': ['*.txt']},
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha'
    ],
)



from tradssat.tmpl.output import OutFile
from tradssat.tmpl.var import FloatVar, IntegerVar


class SoilTempOut(OutFile):
    """
    Reader for DSSAT soil temperature (SOILTEMP.OUT) output files.
    """
    filename = 'SoilTemp.Out'

    def _get_var_info(self):
        return vars_


vars_ = {
    IntegerVar('YEAR', 4, info='Year'),
    IntegerVar('DOY', 3, info='Day of year starting on Jan 1.'),
    IntegerVar('DAS', 5, info='Day after start'),
    FloatVar('TS0D', 7, 1, info='temperature at the surface, oC'),
    FloatVar('TS1D', 7, 1, info='Temperature at soil depth 0-5 cm, oC'),
    FloatVar('TS2D', 7, 1, info='Temperature at soil depth 5-15 cm, oC'),
    FloatVar('TS3D', 7, 1, info='Temperature at soil depth 15-30 cm, oC'),
    FloatVar('TS4D', 7, 1, info='Temperature at soil depth 30-45 cm, oC'),
    FloatVar('TS5D', 7, 1, info='Temperature at soil depth 45-60 cm, oC'),
    FloatVar('TS6D', 7, 1, info='Temperature at soil depth 60-90 cm, oC'),
    FloatVar('TS7D', 7, 1, info='Temperature at soil depth 90-110 cm, oC'),
}



import os
import shutil
import tempfile
import unittest

import numpy.testing as npt

from tradssat import DSSATRun, DSSATResults, set_dssat_dir, ExpFile
from tradssat.exper.exper_vars import TRT_HEAD
from tradssat.out import PlantGroOut

_test_vars = {
    'SOL': 'SRGF',
    'WTH': 'SRAD',
    'EXP': 'FACD',
    'CUL': 'P1',
    'ECO': 'TOPT',
}


class TestRunInput(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.temp_dir = tempfile.TemporaryDirectory()

        mock_dir = shutil.copytree('rsrc/mock_DSSAT', os.path.join(cls.temp_dir.name, 'DSSAT47'))
        set_dssat_dir(mock_dir)
        cls.file = 'rsrc/mock_DSSAT/Exper/Maize/BRPI0202.MZX'

        cls.dssat_run = DSSATRun(cls.file)
        cls.ref_expfile = ExpFile(cls.file)

    def _get_run(self):
        return DSSATRun(self.file)

    def test_get_general_val(self):
        val = self.dssat_run.get_general_val('PEOPLE')
        ref_val = self.ref_expfile.get_value('PEOPLE')

        self.assertEqual(val, ref_val)

    def test_set_general_val(self):
        run = self._get_run()
        run.set_general_val('PEOPLE', 'me')
        new_val = run.get_general_val('PEOPLE')
        self.assertEqual(new_val, 'me')

    def test_n_factor_levels(self):
        ref_n = self.ref_expfile.get_value('INGENO').size

        n = self.dssat_run.n_factor_levels('CULTIVARS')
        self.assertEqual(n, ref_n)

    def test_get_trt_num(self):
        ref_names = self.ref_expfile.get_value('TNAME')
        ref_i_s = self.ref_expfile.get_value('N', sect=TRT_HEAD)

        ref_i = 4
        name = ref_names[ref_i_s == ref_i]
        i = self.dssat_run.get_trt_num(name)

        self.assertEqual(ref_i, i)

    def test_get_trt_name(self):
        ref_names = self.ref_expfile.get_value('TNAME')
        ref_i = self.ref_expfile.get_value('N', sect=TRT_HEAD)

        i = 2
        name = self.dssat_run.get_trt_name(i)

        self.assertEqual(ref_names[ref_i == i], name)

    def test_get_trt_val(self):

        for subfile, vr in _test_vars.items():
            with self.subTest(file=subfile):
                self.dssat_run.get_trt_val(vr, trt=1)

    def test_set_trt_val(self):
        run = self._get_run()

        trt = 5
        val = 15
        run.set_trt_val('IDEP', val, trt=trt)

        new_val = run.get_trt_val('IDEP', trt)

        self.assertEqual(val, new_val)

    def test_add_treatment(self):
        run = self._get_run()
        name = 'Test treatment'

        run.add_treatment(name)

        trts = run.treatments(name=True)
        self.assertIn(name, trts)

    def test_remove_treatment(self):
        run = self._get_run()

        i = 3
        trts = run.treatments(name=True).tolist()
        name = trts.pop(i)

        run.remove_treatment(name)

        new_trts = run.treatments(name=True).tolist()
        self.assertLessEqual(new_trts, trts)

    def test_add_factor_level(self):
        run = self._get_run()

        n_levels = run.n_factor_levels('SA')
        vals = {'SADAT': 2073, 'SANAME': 'New level', 'SABL': 30}
        run.add_factor_level('SA', vals=vals)

        for vr, vl in vals.items():
            self.assertEqual(run.get_factor_level_val(vr, n_levels + 1), vl)

    def test_get_trt_factor_level(self):

        factor = 'CU'
        trt = 5

        factors = self.ref_expfile.get_value(factor)
        trts = self.ref_expfile.get_value('N', sect=TRT_HEAD)

        ref = factors[trts == trt][0]

        val = self.dssat_run.get_trt_factor_level(trt, 'CU')
        self.assertEqual(val, ref)

    def test_set_trt_factor_level(self):
        run = self._get_run()

        new_level = 2
        trt = 1
        run.set_trt_factor_level(trt=trt, factor='CU', level=new_level)

        ref_val = run.get_factor_level_val('INGENO', new_level)
        new_val = run.get_trt_val('INGENO', trt)

        self.assertEqual(ref_val, new_val)

    def test_get_factor_level_val(self):
        for subfile, vr in _test_vars.items():
            with self.subTest(file=subfile):
                self.dssat_run.get_factor_level_val(vr, level=1)

    def test_set_factor_level_val(self):
        run = self._get_run()
        new_name = 'NEW NAME'
        lvl = 1
        run.set_factor_level_val('CNAME', new_name, level=lvl)
        new_val = run.get_factor_level_val('CNAME', lvl)

        self.assertEqual(new_val, new_name)

    @unittest.skip('not ready')
    def test_clean(self):
        pass

    @unittest.skip('not ready')
    def test_check(self):
        pass

    @unittest.skip('not ready')
    def test_write(self):
        pass

    @unittest.skip('not ready')
    def test_write_overwrite(self):
        pass

    @classmethod
    def tearDownClass(cls):
        cls.temp_dir.cleanup()


class TestRunOutput(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.dir_ = 'rsrc/mock_DSSAT/Out'
        cls.result = DSSATResults(cls.dir_)

    def test_get_value(self):
        var = 'RL1D'
        trt = 1
        ref_val = PlantGroOut(os.path.join(self.dir_, 'PlantGro.OUT')).get_value(var, sect={'TREATMENT': trt})
        val = self.result.get_value(var, trt=trt)

        npt.assert_equal(val, ref_val)

    def test_get_value_at_time(self):
        var = 'BWAD'
        trt = 1

        l_at = {'YEAR DOY': '1989 301', 'DAS': 142, 'DAP': 135}

        vec_t = self.result.get_value('DAS', trt)
        vec_vals = self.result.get_value(var, trt)
        ref_val = vec_vals[vec_t == l_at['DAS']]

        for at, t in l_at.items():
            with self.subTest(at):
                val = self.result.get_value(var, t=t, at=at, trt=trt)
                npt.assert_equal(ref_val, val)

    def test_get_final_value(self):
        d_vars = {'FinalOut': {'var': 'HWAM', 'ref_val': 3136},
                  'Out': {'var': 'FWAD', 'ref_val': 3136},
                  'Out&FinalOut': {'var': 'NUCM', 'ref_val': 118.36}}
        trt = 1
        for test, info in d_vars.items():
            with self.subTest(test):
                val = self.result.get_final_value(info['var'], trt=trt)
                self.assertAlmostEqual(info['ref_val'], val, places=0)



from tradssat.tmpl.var import CharacterVar, FloatVar

header_vars = [
    CharacterVar('SLSOURCE', 11, spc=2, info='Source'),
    CharacterVar('SLTX', 5, info='Texture, code'),
    FloatVar('SLDP', 5, 0, info='Depth, cm'),
    CharacterVar('SLDESCRIP', 50, info='Description or local classification')
]

main_vars = {
    CharacterVar('SITE', 11, info='Site name'),
    CharacterVar('COUNTRY', 12, info='Country name'),
    FloatVar('LAT', 8, 3, info='Latitude'),
    FloatVar('LONG', 7, 3, info='Longitude'),
    CharacterVar('SCS FAMILY', 50, info='Family, SCS system'),

    CharacterVar('SCOM', 5, info='Color, moist, Munsell hue'),
    FloatVar('SALB', 5, 2, info='Albedo, fraction '),
    FloatVar('SLU1', 5, 1, info='Evaporation limit, cm'),
    FloatVar('SLDR', 5, 2, info='Drainage rate, fraction day-1'),
    FloatVar('SLRO', 5, 1, info='Runoff curve number (Soil Conservation Service)'),
    FloatVar('SLNF', 5, 2, info='Mineralization factor, 0 to 1 scale'),
    FloatVar('SLPF', 5, 2, info='Photosynthesis factor, 0 to 1 scale'),
    CharacterVar('SMHB', 5, info='pH in buffer determination method, code'),
    CharacterVar('SMPX', 5, info='Phosphorus, extractable, determination code'),
    CharacterVar('SMKE', 5, info='Potassium determination method, code'),

    FloatVar('SLB', 5, 0, info='Depth, base of layer, cm'),
    CharacterVar('SLMH', 5, info='Master horizon'),
    FloatVar('SLLL', 5, 3, info='Lower limit, cm3 cm-3'),
    FloatVar('SDUL', 5, 3, info='Upper limit, drained, cm3 cm-3'),
    FloatVar('SSAT', 5, 3, info='Upper limit, saturated, cm3 cm-3'),
    FloatVar('SRGF', 5, 3, info='Root growth factor, 0.0 to 1.0 '),
    FloatVar('SSKS', 5, 1, info='Sat. hydraulic conductivity, macropore, cm h-1'),
    FloatVar('SBDM', 5, 2, info='Bulk density, moist, g cm-3'),
    FloatVar('SLOC', 5, 2, info='Organic carbon, %'),
    FloatVar('SLCL', 5, 1, info='Clay (<0.002 mm), %'),
    FloatVar('SLSI', 5, 1, info='Silt (0.05 to 0.002 mm), %'),
    FloatVar('SLCF', 5, 1, info='Coarse fraction (>2 mm), %'),
    FloatVar('SLNI', 5, 2, info='Total nitrogen, %'),
    FloatVar('SLHW', 5, 1, info='pH in water'),
    FloatVar('SLHB', 5, 2, info='pH in buffer'),
    FloatVar('SCEC', 5, 1, info='Cation exchange capacity, cmol kg-1'),
    FloatVar('SADC', 5, 1, info='')
}
